/******************************************************************************
Copyright (c) Microsoft Corporation.
Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.
THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
function __decorate(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
}
function __metadata(metadataKey, metadataValue) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(metadataKey, metadataValue);
}

/**
 * Filter object with filter function `fn`.
 *
 * @param obj Object to filter.
 * @param fn Filter function receiving [key, value].
 * @returns A new object filtered.
 */ const filter = (obj, fn)=>Object.fromEntries(Object.entries(obj).filter(fn))
;
/**
 * Pick properties from an object into a new object.
 *
 * @param obj Object to pick from.
 * @param props Properties to pick.
 * @returns A new object with the picked properties.
 */ const pick = (obj, props)=>props.reduce((p, n)=>{
        if (n in obj) p[n] = obj[n];
        return p;
    }, {})
;
/**
 * Create a new object with certain properties omitted.
 *
 * @param obj Object to omit from.
 * @param props Properties to omit.
 * @returns A new object with the properties omitted.
 */ const omit = (obj, props)=>filter(obj, includesKey(props, true))
;
 const includesKey = (props, invert = false)=>([key])=>invert ^ props.includes(key)
;

const bool$1 = Symbol();
/**
 * Convert a function with a settings object to fluent API.
 *
 * ```ts
 * import { toFluent, bool } from 'to-fluent'
 *
 * const cb = toFluent(
 *   class {
 *     foo = bool // indicate boolean but initially omitted
 *     bar?: string // optional
 *     zoo = 123 // a default
 *   },
 *   settings => () => settings
 * )
 *
 * expect(cb()).toEqual({ zoo: 123 })
 * expect(cb.foo()).toEqual({ foo: true, zoo: 123 })
 * expect(cb.not.foo()).toEqual({ foo: false, zoo: 123 })
 * expect(cb.bar('hello')()).toEqual({ bar: 'hello', zoo: 132 })
 * expect(cb.foo.bar('hello').zoo(456)())
 *   .toEqual({ foo: true, bar: 'hello', zoo: 456 })
 * ```
 */ const toFluent$1 = (Schema, cb)=>{
    const bools = Object.entries(new Schema()).filter(([, x])=>x === bool$1 || typeof x === 'boolean'
    );
    const flags = bools.map(([key])=>key
    );
    const omitted = bools.filter(([, x])=>x === bool$1
    ).map(([key])=>key
    );
    const settings1 = omit(new Schema(), omitted);
    let not = false;
    const bind = (settings)=>new Proxy(cb, {
            get (_, key, receiver) {
                if (key === 'not') {
                    not = true;
                    return receiver;
                } else if (flags.includes(key)) {
                    const value = not ? false : true;
                    not = false;
                    return bind({
                        ...settings,
                        [key]: value
                    });
                } else {
                    return (value)=>{
                        return bind({
                            ...settings,
                            [key]: value
                        });
                    };
                }
            },
            apply: (_, self, args)=>cb.call(self, {
                    ...settings
                }).apply(self, args)
        })
    ;
    return bind(settings1);
};

const entries = (obj)=>Object.entries(obj);
const fromEntries = (entries1)=>Object.fromEntries(entries1);
const accessors = (target, source, fn, filter = ()=>true)=>{
    const prevDesc = new Map();
    Object.defineProperties(target, Object.fromEntries(entries(source).filter(([key])=>typeof key === 'string').filter(([key, value])=>filter(key, value)).map(([key, value])=>{
        const next = fn(key, value);
        const prev = Object.getOwnPropertyDescriptor(target, key);
        prevDesc.set(key, prev);
        if (prev && prev.get && prev.set) {
            const { get , set  } = next;
            next.get = ()=>{
                // run both getters if any, for side-effects to run
                // the correct behavior here? maybe pass the previous
                // getter as a param to the next one and let them decide?
                // but it works.
                const top = get?.();
                const below = prev.get();
                return below ?? top;
            };
            next.set = (v)=>{
                prev.set?.(v);
                set(prev.get?.() ?? v);
            };
        }
        //!? 'create accessor', key, prev, next
        return [
            key,
            {
                configurable: true,
                enumerable: true,
                ...next
            }
        ];
    })));
    return ()=>{
        for (const [key, desc] of prevDesc){
            if (desc == null) {
                delete target[key];
            } else {
                // const newValue = (target as any)[key]
                Object.defineProperty(target, key, desc);
            }
        }
    };
};
const kebab = (s)=>s.replace(/[a-z](?=[A-Z])|[A-Z]+(?=[A-Z][a-z])/g, '$&-').toLowerCase();
const styleToCss = (style)=>{
    let css = '';
    for(const key in style)css += style[key] ? kebab(key) + ':' + style[key] + ';' : '';
    return css;
};
const removeFromArray = (arr, el, quiet = false)=>{
    const index = arr.indexOf(el);
    if (!~index) {
        if (quiet) return;
        throw new ReferenceError('Item to be removed does not exist in the array.');
    }
    return arr.splice(index, 1);
};
const shallowEqual = (a1, b1)=>[
        [
            a1,
            b1
        ],
        [
            b1,
            a1
        ]
    ].every(([a, b])=>entries(a).every(([key, value])=>key in b && b[key] == value));
const getOwnProperty = (object, name)=>Object.getOwnPropertyDescriptor(object, name)?.value;
const defineProperty = toFluent$1(class {
    configurable = false;
    enumerable = false;
    writable = bool$1;
}, (descriptor)=>(object, name, value)=>Object.defineProperty(object, name, value == null ? descriptor : {
            ...descriptor,
            value
        }));
const filterMap = (array, fn)=>array.map(fn).filter(Boolean);

const isEvent = (name)=>name.charAt(0) === 'o' && name.charAt(1) === 'n' && !name.endsWith('ref');
const toEvent = (name)=>name.slice(2);
const createProp = (doc = html, el, _type, name, value, attrs)=>{
    // all the Any's below are on purpose
    // all the cases should be taken care of,
    // but since this is user input they are allowed
    // to do something wrong and we want to raise an
    // special cases
    switch(name){
        case 'children':
            return;
        case 'style':
            // if we createAttribute and set .value then that
            // triggers the css parser and we can't compare if
            // doing it this way retains the exact string and
            value = value?.valueOf();
            if (typeof value === 'object') value = styleToCss(value);
            if (value) {
                el.setAttribute('style', value);
                // we know there is an attribute node because we
                attrs.style = el.getAttributeNode('style');
            }
            return;
    }
    //
    // create prop
    // we used to normalize the attribute name
    // but now we don't. because there is a high
    // normalizes attribute names we leave it here
    const attr = name // toAttr[name] || name
    ;
    value = value?.valueOf();
    switch(typeof value){
        case 'string':
        case 'number':
            if (doc === svg || !(name in el)) {
                el.setAttribute(attr, value);
                attrs[attr] = el.getAttributeNode(attr);
                return;
            }
            break;
        case 'function':
            if (isEvent(attr)) {
                //!? 'create event listener', attr, value
                el.addEventListener(toEvent(attr), value, value);
            } else {
                el.setAttribute(attr, '');
                attrs[attr] = el.getAttributeNode(attr);
                el[name] = value;
            }
            return;
    }
    if (value == null) {
        el.removeAttribute(attr);
        delete attrs[attr];
        if (el[name] != null) {
            delete el[name];
        }
        return;
    }
    el[name] = value;
    if (el.hasAttribute(attr)) attrs[attr] = el.getAttributeNode(attr);
};
const propCache = new WeakMap();
const createProps = (doc, el, type, props = {}, attrs = {}, cacheRef = el)=>{
    for(const name in props)createProp(doc, el, type, name, props[name], attrs);
    propCache.set(cacheRef, {
        props,
        attrs
    });
};
const updateProps = (doc, el, type, next = {}, cacheRef = el)=>{
    if (!propCache.has(cacheRef)) return next && createProps(doc, el, type, next, void 0, cacheRef);
    const c = propCache.get(cacheRef);
    const { attrs , props  } = c;
    if (!next) {
        //!? 'disconnected, removing all attrs and props', cacheRef
        for(const name in attrs){
            el.removeAttributeNode(attrs[name]);
        }
        for(const name1 in props){
            if (isEvent(name1)) {
                //!? 'removed event listener', name
                el.removeEventListener(toEvent(name1), props[name1], props[name1]);
            } else {
                delete el[name1];
            }
        }
        propCache.delete(cacheRef);
        return;
    }
    let prev;
    let value;
    out: for(const name2 in props){
        if (!(name2 in next)) {
            //!? 'removed prop', name
            delete el[name2];
            continue;
        }
        value = next[name2];
        switch(name2){
            case 'children':
                continue out;
        }
        value = value?.valueOf()
        ;
        prev = props[name2];
        if (prev !== value) {
            if (typeof value === 'function') {
                if (prev.fn && value.fn) {
                    //!? 'updated fn'
                    value = prev.update(value.fn, value.options);
                    if (value === prev) {
                        next[name2] = value;
                    }
                }
                if (prev !== value || cacheRef !== el) {
                    let attr = name2 // toAttr[name] || name
                    ;
                    if (isEvent(attr)) {
                        attr = toEvent(attr);
                        //!? 'updated event listener', attr, prev, value
                        el.removeEventListener(attr, prev, prev);
                        el.addEventListener(attr, value, value);
                    } else {
                        props[attr] = el[attr] = value;
                    }
                }
            } else if (!(name2 in attrs)) {
                if (value == null) {
                    //!? 'removed prop', name
                    if (el[name2] != null) {
                        el[name2] = value;
                    }
                } else {
                    el[name2] = value;
                }
            }
        }
    }
    for(const name3 in attrs){
        if (!(name3 in next) || next[name3] === false || next[name3] == null) {
            //!? 'removed attr', name
            el.removeAttribute(name3)
            ;
            delete attrs[name3];
            continue;
        }
        value = next[name3]?.valueOf();
        switch(name3){
            case 'style':
                if (typeof value === 'object') value = styleToCss(value) //|| null
                ;
                break;
        }
        // updated value
        if (props[name3] !== value && typeof value !== 'function') {
            if (value == null) {
                if (attrs[name3].value != null) {
                    //!? 'removed attr', name
                    el.removeAttribute(name3);
                    delete attrs[name3];
                }
            } else {
                //!? 'updated attr', name, value
                attrs[name3].value = value;
            }
        }
    }
    for(const name4 in next){
        if (!(name4 in attrs) && !(name4 in props)) createProp(doc, el, type, name4, next[name4], attrs);
    }
    c.props = next;
};

const anchor = new Comment();
const Fragment = Symbol();
const jsx = (kind, props, key)=>kind === Fragment ? props.children : {
        kind,
        props,
        key
    };
let hook;
const createHook = ()=>function current(fn = current.fn) {
        const prev = hook;
        hook = current;
        hook.fn = fn;
        fn();
        hook = prev;
    };
let Chunk = class Chunk extends Array {
    firstChild;
    dom = [];
    get last() {
        return this.dom.at(-1);
    }
    get nextSibling() {
        return this.last?.nextSibling;
    }
    //   this.unshift(x)
    appendChild(x) {
        this.push(x);
    }
    after(x) {
        this.last?.after(x);
    }
    save() {
        this.dom = [
            ...this
        ];
    }
    remove() {
        this.dom.forEach((el)=>{
            el.onunref?.();
            el.remove();
        });
        this.splice(0);
    }
    removeChild(x) {
        const i = this.indexOf(x);
        ~i && this.splice(i, 1);
    // i = this.dom.indexOf(x)
    // ~i && this.dom.splice(i, 1)
    }
};
const { TEXT_NODE , COMMENT_NODE  } = document;
const html = document.createElement.bind(document);
const svg = document.createElementNS.bind(document, 'http://www.w3.org/2000/svg');
const forceArray = (x, withNull)=>Array.isArray(x) ? withNull && !x.length ? [
        null
    ] : x : x == null && !withNull ? [] : [
        x
    ];
const flatDom = (arr, res = [])=>{
    for (const el of arr){
        if (el.dom) res.push(...flatDom(el.dom));
        else res.push(el);
    }
    return res;
};
const prevs = new WeakMap();
function render(n, el = document.createDocumentFragment(), doc = html, withNull = false) {
    reconcile(el, forceArray(n, withNull), prevs.get(el), doc);
    return el;
}
const reconcile = (parent, nk, pk, doc)=>{
    if (pk?.running) {
        console.warn('attempt to reconcile lane which is reconciling');
        return;
    }
    if (pk === nk) nk = [
        ...nk
    ];
    for (const [i, n] of nk.entries()){
        nk[i] = n?.valueOf?.();
    }
    prevs.set(parent, nk);
    nk.running = true;
    nk.dom = Array(nk.length);
    nk.keyed = new Map();
    nk.mapped = new Map();
    if (Array.isArray(pk)) {
        const keep = new Set();
        for(let i1 = 0, n1, el, p, pel, k, pi; i1 < nk.length; i1++){
            n1 = nk[pi = i1];
            k = n1?.key;
            if (k != null) {
                nk.keyed.set(k, i1);
                pi = pk.keyed.get(k) ?? -1;
            }
            p = pk[pi];
            pel = pk.dom[pi];
            nk.dom[i1] = el = create(doc, n1, p, pel);
            if (el === pel) keep.add(pel);
        }
        for (const pel1 of pk.dom){
            if (!keep.has(pel1)) {
                pel1.onunref?.();
                if (parent instanceof Chunk) parent.removeChild(pel1);
                pel1.remove();
                pk.mapped.get(pel1)?.hook?.onremove?.();
            }
        }
    } else {
        for(let i2 = 0, n2, el1, k1; i2 < nk.length; i2++){
            n2 = nk[i2];
            k1 = n2?.key;
            if (k1 != null) nk.keyed.set(k1, i2);
            nk.dom[i2] = el1 = create(doc, n2);
            nk.mapped.set(el1, n2);
        }
    }
    nk.flatDom = flatDom(nk.dom);
    if (pk?.flatDom) diff(parent, nk.flatDom, pk.flatDom);
    else nk.flatDom.forEach((el)=>parent.appendChild(el));
    nk.running = false;
};
// scheduling to prevent triggering ref effects before this render finishes
// TODO: has to be put in a proper queue instead of relying on the microtask
const mount = (el)=>{
    queueMicrotask(()=>{
        if (el?.ref && el !== el.ref.current) {
            el.ref.current = el;
        }
        if (el?.onref) {
            el.onunref = el.onref(el);
        }
    });
    return el;
};
const diff = (parent, n, p, i = 0, len = n.length, el, last)=>{
    if (parent instanceof Chunk) {
        // TODO: optimize this
        for(; i < len; i++){
            el = n[i];
            if (i < parent.length) {
                if (p[i] === el) continue;
                parent[i] = el;
            } else {
                parent.push(el);
            }
        }
        let d = parent.length - len;
        while(d--)parent.pop();
    } else {
        for(; i < len; i++){
            el = n[i];
            if (p[i] === el) last = el;
            else if (!i) parent.insertBefore(last = el, parent.firstChild);
            else last.after(last = el);
        }
    }
};
const create = (doc, n, p, pel)=>{
    let el;
    switch(typeof n){
        case 'string':
        case 'number':
            if (pel?.nodeType === TEXT_NODE) {
                if (p != n) pel.nodeValue = n;
                return pel;
            }
            el = new Text(n);
            return el;
        case 'object':
            if (!n) break;
            if (Array.isArray(n)) {
                if (pel && Array.isArray(p)) el = pel;
                else el = new Chunk();
                if (!n.length) n.push(null);
                reconcile(el, n, p, doc);
                el.save();
            } else if (typeof n.kind === 'string') {
                // maybe switch to svg namespace
                if (n.kind === 'svg') doc = svg;
                if (n.props.ref?.current && n.props.ref.current.tagName.toLowerCase() === n.kind && (el = n.props.ref.current) || // if the previous element is of the same kind, use that
                pel && p?.kind === n.kind && (el = pel)) {
                    updateProps(doc, el, n.kind, n.props);
                } else {
                    // otherwise create an element and props
                    el = doc(n.kind, 'is' in n.props ? {
                        is: n.props.is
                    } : void 0);
                    createProps(doc, el, n.kind, n.props);
                }
                if (n.kind === 'foreignObject') doc = html;
                // render children
                render(n.props.children, el, doc);
                mount(el);
            } else {
                let initial = true;
                if (!((el = pel) && (n.hook = p?.hook))) {
                    el = new Chunk();
                    n.hook = createHook();
                }
                let prevDom;
                let nextDom;
                n.hook(()=>{
                    let next;
                    if (!initial && !(next = el.nextSibling)) el.after(next = anchor);
                    render(n.kind(n.props), el, doc, true);
                    el.save();
                    if (!initial && next) {
                        nextDom = flatDom(el);
                        if (prevDom?.length > 0) {
                            for(let i = 0, e; i < nextDom.length; i++){
                                e = nextDom[i];
                                if (prevDom[i] !== e) next.before(e);
                            }
                        } else {
                            for (const e1 of nextDom)next.before(e1);
                        }
                        prevDom = nextDom;
                        next === anchor && next.remove();
                    } else {
                        initial = false;
                    }
                });
            }
            return el;
    }
    if (pel?.nodeType === COMMENT_NODE) return pel;
    el = new Comment();
    return el;
};

/**
 * Creates a proxy that calls `cb` for every property accessed and returns its result.
 *
 * @param cb Called with the key of the property being accessed.
 */ const Getter$1 = (cb)=>new Proxy({}, {
        get: (_, key)=>cb(key)
    });
const FluentCaptureSymbol = Symbol();
const FluentCapture = ()=>{
    const fn = ()=>{};
    const results = [];
    const proxy = new Proxy(fn, {
        get: (_, key, receiver)=>{
            if (key === FluentCaptureSymbol) return true;
            if (key === '_results') return results;
            results.push([
                'get',
                key
            ]);
            return receiver;
        },
        apply: (_, __, args)=>{
            results.push([
                'apply',
                args
            ]);
            return proxy;
        }
    });
    return proxy;
};
const applyFluentCapture = (capture, target)=>{
    let res = target;
    for (const [op, args] of capture){
        if (op === 'get') {
            res = res[args];
        } else if (op === 'apply') {
            res = res(...args);
        }
    }
    return res;
};

const argsOf = (x)=>{
    const d = x.indexOf('{') + 1;
    if (!d || x.slice(0, d).includes(')')) return [];
    x = x.slice(d);
    let match;
    do {
        match = false;
        x = x.replace(/`((\\`)?[^`])+?`|'((\\')?[^'])+?'|"((\\")?[^"])+?"|\([^{[(]*?\)|\{[^{[(]*?\}|\[[^{[(]*?\]/g, ()=>{
            match = true;
            return '';
        });
    }while (match)
    return x.split('}')[0].split(',').map((x)=>x.split(/[^\w\s$]+/g)[0].trim());
};
/**
 * Extract destructured argument names from `fn`'s first argument and return a `Set` of `keys`.
 *
 * ```ts
 * const fn = ({ foo, bar }) => void 0
 * argtor(fn) // => Set(2) {'foo', 'bar'}
 * ```
 *
 * @param fn The function to extract argument keys from.
 * @param keys A keys `Set` to fill in with the argument names.
 * @returns The keys `Set` with the argument names.
 */ const argtor = (fn, keys = new Set())=>(argsOf('' + fn).forEach((x)=>!x || keys.add(x)), keys);

const bool = Symbol.for('to-fluent-bool');
/**
 * Convert a function with a settings object to fluent API.
 *
 * ```ts
 * import { toFluent, bool } from 'to-fluent'
 *
 * const cb = toFluent(
 *   class {
 *     foo = bool // indicate boolean but initially omitted
 *     bar?: string // optional
 *     zoo = 123 // a default
 *   },
 *   settings => () => settings
 * )
 *
 * expect(cb()).toEqual({ zoo: 123 })
 * expect(cb.foo()).toEqual({ foo: true, zoo: 123 })
 * expect(cb.not.foo()).toEqual({ foo: false, zoo: 123 })
 * expect(cb.bar('hello')()).toEqual({ bar: 'hello', zoo: 132 })
 * expect(cb.foo.bar('hello').zoo(456)())
 *   .toEqual({ foo: true, bar: 'hello', zoo: 456 })
 * ```
 */ const toFluent = (Schema, cb)=>{
    const bools = Object.entries(new Schema()).filter(([, x])=>x === bool || typeof x === 'boolean');
    const flags = bools.map(([key])=>key);
    const omitted = bools.filter(([, x])=>x === bool).map(([key])=>key);
    const settings = omit(new Schema(), omitted);
    let not = false;
    const bind = (settings)=>new Proxy(cb, {
            get (_, key, receiver) {
                if (key === 'not') {
                    not = true;
                    return receiver;
                } else if (flags.includes(key)) {
                    const value = not ? false : true;
                    not = false;
                    return bind({
                        ...settings,
                        [key]: value
                    });
                } else {
                    return (value)=>{
                        return bind({
                            ...settings,
                            [key]: value
                        });
                    };
                }
            },
            apply: (_, self, args)=>cb.call(self, {
                    ...settings
                }).apply(self, args)
        });
    return bind(settings);
};

class AbortOptions {
    throw = bool;
    latest = bool;
    timeout;
}
const abort = toFluent(AbortOptions, (options)=>(fn)=>{
        let ctrl;
        return async function(...args) {
            if (options.latest && ctrl) ctrl.abort();
            ctrl = new AbortController();
            if (options.timeout != null) {
                const timeoutError = new Error('Timed out');
                try {
                    return await Promise.race([
                        new Promise((_, reject)=>setTimeout(reject, options.timeout, timeoutError)),
                        fn(ctrl.signal).apply(this, args),
                    ]);
                } catch (error) {
                    if (error === timeoutError) {
                        ctrl.abort();
                        if (options.throw) throw error;
                    } else {
                        if (options.throw) throw error;
                    }
                }
            } else {
                return fn(ctrl.signal).apply(this, args);
            }
        };
    });

const MouseButton = {
    Left: 1,
    Right: 2,
    Middle: 4
};

class DispatchOptions {
    bubbles = bool;
    cancelable = bool;
    composed = bool;
}
const dispatchEvent = (el, nameOrEvent, detail, init, options)=>el.dispatchEvent(nameOrEvent instanceof Event ? nameOrEvent : new CustomEvent(nameOrEvent, {
        detail,
        ...init,
        ...options
    }));
const dispatch = toFluent(DispatchOptions, (options)=>(el, nameOrEvent, detail, init)=>dispatchEvent(el, nameOrEvent, detail, init, options));
const dispatchBind = (el)=>toFluent(DispatchOptions, (options)=>(nameOrEvent, detail, init)=>dispatchEvent(el, nameOrEvent, detail, init, options));

const Task = (fn, self, args)=>{
    let resolve;
    let reject;
    const promise = new Promise((rs, rj)=>{
        resolve = rs;
        reject = rj;
    });
    return {
        promise,
        resolve,
        reject,
        fn,
        self,
        args
    };
};
const runTask = (task, res)=>(task.resolve(res = task.fn.apply(task.self, task.args)), res);
const groupTasks = (other, tasks)=>tasks.forEach((t)=>other.promise.then(t.resolve));

class QueueOptions {
    first = bool;
    last = bool;
    next = bool;
    raf = bool;
    task = bool;
    time = bool;
    atomic = bool;
    debounce;
    throttle;
}
const wrapQueue = (options = {})=>(fn)=>{
        const initialOptions = {
            ...options
        };
        const queued = [];
        let queueFn;
        let id;
        let last;
        let runs = false;
        if (options.raf) queueFn = requestAnimationFrame;
        else if (options.task) queueFn = queueMicrotask;
        else if (options.time) queueFn = setTimeout;
        else if (options.throttle != null) {
            queueFn = (fn)=>{
                setTimeout(fn, options.throttle);
            };
        } else if (options.debounce != null) {
            queueFn = (fn)=>{
                clearTimeout(id);
                id = setTimeout(fn, options.debounce);
            };
        } else if (options.atomic) queueFn = (fn)=>fn();
        else {
            // No queue function provided, return identity.
            // This is used when extending this in `event`.
            return fn;
        }
        if (options.first == null && options.last == null) {
            if (options.throttle != null) options.first = true;
            else if (options.debounce != null) options.last = true;
            else options.next = true;
        }
        const cb = ()=>{
            let task;
            if (queued.length) {
                if (options.atomic) {
                    task = queued.shift();
                    runTask(task).catch((_error)=>{
                    }).finally(()=>queueFn(cb));
                    return;
                }
                if (options.last) {
                    if (options.next) {
                        const left = queued.splice(0, queued.length - 1);
                        task = left.pop() ?? queued.pop();
                        groupTasks(task, left);
                        last = runTask(task);
                        if (queued.length) {
                            queueFn(cb);
                            return;
                        }
                    } else {
                        task = queued.pop();
                        groupTasks(task, queued.splice(0));
                        last = runTask(task);
                    }
                } else if (options.next) {
                    task = queued.shift();
                    groupTasks(task, queued.splice(0, queued.length - 1));
                    queueFn(cb);
                    last = runTask(task);
                    return;
                } else {
                    task = Task();
                    groupTasks(task, queued.splice(0));
                    task.resolve(last);
                }
            }
            runs = false;
        };
        function wrapped(...args) {
            //!? 'wrap called'
            const task = Task(fn, this, args);
            if (!runs && options.first) {
                runs = true;
                last = runTask(task);
                queueFn(cb);
                return task.promise;
            }
            queued.push(task);
            if (!runs || options.debounce) {
                runs = true;
                queueFn(cb);
            }
            return task.promise;
        }
        wrapped.fn = fn;
        wrapped.options = initialOptions;
        wrapped.update = (newFn, newOptions)=>{
            //!? 'updating fn'
            if (!shallowEqual(initialOptions, newOptions)) {
                //!? 'new options', initialOptions, newOptions
                return newFn;
            }
            fn = newFn;
            return wrapped;
        };
        return wrapped;
    };
/**
 * Decorate function `fn` with a queue function.
 *
 * The decorated function will returns a `Promise` that resolves with its result.
 * All calls will be resolved with latest result at any given quantum.
 *
 * Options:
 * - `first` => Run only first, then debounce.
 * - `last` => Run last (default behavior when using `debounce`).
 * - `next` => Run final subsequent call on next quantum (default behavior when nothing is set and not a `debounce`).
 *
 * - `raf` => Queue with `requestAnimationFrame`.
 * - `task` => Queue with `queueMicrotask`.
 * - `time` => Queue with `setTimeout`.
 *
 * - `debounce(ms)` => Debounce at specified `ms`.
 * - `throttle(ms)` => Throttle at specified `ms`.
 *
 * ```ts
 * fn = (x: number) => console.log(x)
 *
 * const cb = queue.task((x: number) => count += x)
 * cb(1) // passes (after task)
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // passes
 *
 * const cb = queue.task.last((x: number) => count += x)
 * cb(1) // discarded
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // passes
 *
 * const cb = queue.task.first((x: number) => count += x)
 * cb(1) // passes (before task)
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // discarded
 *
 * const cb = queue.task.first.last((x: number) => count += x)
 * cb(1) // passes (before task)
 * cb(2) // discarded
 * cb(3) // discarded
 * cb(4) // passes
 *
 * const cb = queue.task.first.last.next((x: number) => count += x)
 * cb(1) // passes (before task)
 * cb(2) // discarded
 * cb(3) // passes (after task)
 * cb(4) // passes (next task)
 *
 * const cb = queue.task.last.next((x: number) => count += x)
 * cb(1) // discarded
 * cb(2) // discarded
 * cb(3) // passes (after task)
 * cb(4) // passes (next task)
 * ```
 */ const queue = toFluent(QueueOptions, wrapQueue);

class EventOptions extends QueueOptions {
    prevent = bool;
    stop = bool;
    immediate = bool;
    capture = bool;
    once = bool;
    passive = bool;
}
const wrapEvent = (options = {})=>(fn = ()=>{})=>wrapQueue(options)(options.prevent || options.stop || options.immediate || options.capture != null || options.once != null || options.passive != null ? Object.assign(function(e) {
            if (options.prevent) e.preventDefault();
            if (options.stop) {
                options.immediate ? e.stopImmediatePropagation() : e.stopPropagation();
            }
            return fn.call(this, e);
        }, options) : fn);
/**
 * Decorates event handler `fn`.
 *
 * Options:
 * - `prevent` => `event.preventDefault()`
 * - `stop` => `event.stopPropagation()`
 * - `stop.immediate` => `event.stopImmediatePropagation()`
 *
 * Listener options:
 * - `capture` => If specified, indicates that events of this type will be dispatched to the registered listener before being dispatched to any `EventTarget` beneath it in the DOM tree.
 * - `once` => If specified, indicates that the listener should be invoked at most once after being added. The listener would be automatically removed when invoked.
 * - `passive` => Indicates that the function specified by listener will never call `preventDefault()`. If a passive listener does call `preventDefault()`, the user agent will do nothing other than generate a console warning. If not specified, defaults to `not.passive` – except that in browsers other than Safari and Internet Explorer, defaults to `passive` for the `wheel`, `mousewheel`, `touchstart` and `touchmove` events.
 *
 * Queue options:
 * - `first` => Run only first, then debounce.
 * - `last` => Run last (default behavior when using `debounce`).
 * - `next` => Run final subsequent call on next quantum (default behavior when nothing is set and not a `debounce`).
 *
 * - `raf` => Queue with `requestAnimationFrame`.
 * - `task` => Queue with `queueMicrotask`.
 * - `time` => Queue with `setTimeout`.
 *
 * - `debounce(ms)` => Debounce at specified `ms`.
 * - `throttle(ms)` => Throttle at specified `ms`.
 *
 * ```ts
 * btn.onclick = event(fn)
 * btn.onclick = event.prevent(fn)
 * btn.onclick = event.prevent.stop(fn)
 * btn.onclick = event.stop.immediate(fn)
 *
 * // examples with queue
 * btn.onclick = event.stop.raf(fn)
 * btn.onclick = event.prevent.throttle(50)(fn)
 * btn.onclick = event.debounce(100)(fn)
 * ```
 */ const event = toFluent(EventOptions, wrapEvent);

function chain(first, ...rest) {
    if (Array.isArray(first)) {
        [first, ...rest] = first.flat().filter(Boolean);
        if (first == null) return;
    }
    return (...args)=>{
        first(...args);
        for (const fn of rest)fn(...args);
    };
}
const onAll = (target, listener, ...args)=>{
    const targetOwnDispatch = target.dispatchEvent;
    const events = new Set();
    target.dispatchEvent = function(event) {
        if (!events.has(event.type)) {
            target.addEventListener(event.type, listener, ...args);
            events.add(event.type);
        }
        return targetOwnDispatch.call(this, event);
    };
    return ()=>{
        for (const eventType of events){
            target.removeEventListener(eventType, listener, ...args);
        }
        events.clear();
        target.dispatchEvent = targetOwnDispatch;
    };
};

/**
 * Creates a proxy that calls `cb` for every property accessed and returns its result.
 *
 * @param cb Called with the key of the property being accessed.
 */ const Getter = (cb)=>new Proxy({}, {
        get: (_, key)=>cb(key)
    });

class OnOptions extends EventOptions {
    once = bool;
    passive = bool;
    capture = bool;
}
const onEvent = (el, type, listener, options)=>{
    el.addEventListener(type, listener, options);
    return ()=>el.removeEventListener(type, listener, options);
};
const onEventFluent = (el, key)=>toFluent(OnOptions, (options)=>(listener = ()=>{})=>onEvent(el, key, wrapEvent(options)(listener), options));
function on(el, key) {
    return key != null ? onEventFluent(el, key) : Getter((key)=>onEventFluent(el, key));
}

const NO_DEPS = Symbol();
const REDUCER = Symbol();
class EffectOptions extends QueueOptions {
    cb;
    keys;
    once = bool;
    lock = bool;
    isFulfillReducer = bool;
}
class Fx {
    remove;
    target;
    values = {};
    pass;
    constructor(fx){
        Object.assign(this, fx);
    }
}
let GlobalLock;
class ContextClass {
    entries;
    mem;
    keys;
    effects;
    triggered;
    active;
    reducer;
    $;
    scheduled;
    listeners;
    static attach(target, extension = {}) {
        if (target.context instanceof ContextClass) {
            // TODO: what should be the behavior here?
            return;
        }
        target.context = new ContextClass(target);
        // We are using mutate() to extend $ because at this point there are reactive effects
        // that will trigger and depend on the $ object to be fulfilled, along with its extension.
        // in the DOM context, the extension contains DOM initializers that don't exist in Workers
        target.context.mutate(()=>{
            target.$ = Object.assign(target.context.$, extension);
        });
    }
    constructor(target){
        this.target = target;
        this.keys = new Set();
        this.triggered = new Set();
        this.$ = {};
        this.scheduled = new Set();
        this.listeners = new Map();
        this.maybeAddChangeListener = (key, value)=>{
            if (typeof value !== 'object' || !(value instanceof EventTarget) || typeof Element !== 'undefined' && value instanceof Element) {
                return;
            }
            const [prev, off] = this.listeners.get(key) ?? [];
            if (prev === value) return;
            //!? 'change listener', key
            off?.();
            const ownEvents = new WeakSet();
            const runKeyEffects = ()=>{
                // clearing values so they retrigger for same reference
                this.clearKey(key)
                ;
                this.target[key] = value;
                // and then manually
                this.runEffects(key);
            };
            this.listeners.set(key, [
                value,
                onAll(value, (event)=>{
                    if (event.type === 'change') runKeyEffects();
                    if (ownEvents.has(event)) return;
                    // TODO: better copy event function
                    const newEvent = new event.constructor(event.type, {
                        ...event,
                        detail: event.detail,
                        composed: true,
                        bubbles: true
                    });
                    ownEvents.add(newEvent);
                    if (typeof Element === 'undefined' || !(this.target instanceof Element)) {
                        console.error(this.target);
                        throw new Error('Target is not an EventTarget');
                    }
                    this.target.dispatchEvent(newEvent);
                }),
            ]);
            return true;
        };
        this.cleanup = ()=>{
            if (this.effects) {
                for (const fns of this.effects.values())fns.forEach((f)=>f.dispose?.());
            }
            this.effects = new Map([
                [
                    NO_DEPS,
                    []
                ],
                ...this.entries.map(([key])=>[
                        key,
                        []
                    ]),
            ]);
        };
        this.register = (f)=>{
            f.keys = f.options?.keys ?? argtor(f.fn);
            f.fn = wrapQueue(f.options)(f.fn);
            f.keys.forEach((key)=>{
                const fx = this.effects.get(key);
                if (!fx) {
                    console.warn('No effects for key:', key);
                    return;
                }
                fx.push(f);
            });
            !f.keys.size && this.effects.get(NO_DEPS).push(f);
            f.remove = async ()=>{
                if (f.keys.size) {
                    f.keys.forEach((key)=>{
                        const fx = this.effects.get(key);
                        if (!fx) {
                            console.warn('No effects for key:', key);
                            return;
                        }
                        removeFromArray(fx, f, true);
                    });
                } else removeFromArray(this.effects.get(NO_DEPS), f);
                if (this.triggered.has(f)) this.triggered.delete(f);
                f.dispose?.();
            };
            return this.run(f);
        };
        this.flush = ()=>{
            const current = [
                ...this.triggered
            ];
            this.triggered.clear();
            current.forEach(this.run);
        };
        this.run = (f)=>{
            if (f.pass && f.options?.once) return;
            if (this.active || GlobalLock) {
                if (f !== this.active) this.triggered.add(f);
                GlobalLock?.add(this);
                return;
            }
            this.active = f;
            if (this.update(f)) {
                f.dispose?.();
                if (!f.pass) {
                    f.values = {};
                    return this.complete();
                }
            }
            return this.finalize();
        };
        this.complete = (result)=>{
            const f = this.active;
            if (!f.pass) {
                this.active = null;
                this.flush();
                return;
            }
            if (f.cb?.(result) === false) {
                this.active = null;
                this.flush();
                return;
            }
            if (f.target != null && !f.options?.isFulfillReducer) {
                this.target[f.target] = result;
                this.active = null;
                this.flush();
                return true;
            } else {
                if (typeof result === 'function') {
                    f.dispose = ()=>{
                        result();
                        f.dispose = null;
                    };
                }
                this.active = null;
                this.flush();
            }
        };
        this.reduce = toFluent(EffectOptions, (options)=>(fn, initial)=>{
                this.reducer = new Fx({
                    fn,
                    options,
                    initial
                });
                return REDUCER;
            });
        this.fulfill = toFluent(EffectOptions, (options)=>(fn, initial)=>{
                options.isFulfillReducer = true;
                this.reducer = new Fx({
                    fn,
                    options,
                    initial
                });
                return REDUCER;
            });
        this.effect = toFluent(EffectOptions, (options)=>(fn, cb)=>{
                const f = new Fx({
                    fn,
                    cb: cb ?? options.cb,
                    options
                });
                this.register(f);
                return f.remove;
            });
        this.mutate = (fn)=>{
            const f = new Fx({
                fn,
                options: {
                    lock: true
                }
            });
            f.keys = new Set();
            f.values = {
                ...this.target
            } //as Deps<T>
            ;
            return this.run(f);
        };
        this.mutating = (fn)=>()=>{
                this.mutate(fn);
            };
        this.use = (fn)=>fn(this.target);
        this.using = (fn)=>()=>fn(this.target);
        this.function = (fn)=>{
            const { withLock , target  } = this;
            function wrapped(...args) {
                let result;
                withLock(()=>{
                    const cb = fn(target);
                    result = cb.apply(this, args);
                });
                return result;
            }
            return wrapped;
        };
        this.with = (fn)=>{
            const { target  } = this;
            function wrapped(...args) {
                const cb = fn(target);
                return cb.apply(this, args);
            }
            return wrapped;
        };
        this.callback = toFluent(EffectOptions, (options)=>(fn)=>{
                const { mutate  } = this;
                const qfn = wrapQueue(options)(function wrapped(...args) {
                    mutate((ctx)=>{
                        const cb = fn(ctx);
                        cb.apply(this, args);
                    });
                });
                return qfn;
            });
        this.atomic = (fn)=>this.callback(()=>fn);
        this.query = (sel)=>this.reduce(({ root  })=>root.querySelector(sel));
        this.when = (condition, cb)=>function(...args) {
                if (condition) cb.apply(this, args);
            };
        this.withGlobalLock = (fn)=>{
            GlobalLock = new Set();
            try {
                fn();
            } catch (error) {
                console.warn(error);
            }
            const items = [
                ...GlobalLock
            ];
            GlobalLock = null;
            for (const item of items){
                item.flush();
            }
        };
        this.withLock = (fn)=>{
            this.active = fn;
            try {
                fn();
            } catch (error) {
                console.warn(error);
            }
            this.active = null;
            this.flush();
        };
        this.copyMethods();
        this.createMemory();
        this.createAccessors();
        this.cleanup();
        for (const item of this.scheduled){
            if (item.value) {
                const { key , value: [method, args]  } = item;
                this.$[key] = this.$[method](...args);
            } else if (item.captured) {
                const { key: key1 , captured  } = item;
                this.$[key1] = applyFluentCapture(captured, this.$);
            }
        }
    }
    copyMethods() {
        Object.assign(this.$, pick(this, [
            'atomic',
            'callback',
            'cleanup',
            'effect',
            'fulfill',
            'function',
            'mutate',
            'mutating',
            'reduce',
            'register',
            'render',
            'use',
            'using',
            'query',
            'when',
            'with',
            'withLock',
            'globalLock',
        ]));
    }
    runEffects(key) {
        this.effects.get(key).forEach(this.run);
    }
    clearKey(key) {
        this.effects.get(key).forEach((f)=>{
            delete f.values[key];
        });
    }
    // executeReducersFor(targetKey: keyof T) {
    //   for (const fx of this.effects.values()) {
    //       if (f.target === targetKey) {
    //       }
    //   }
    maybeAddChangeListener;
    createMemory() {
        const pending = this.target._pendingProperties;
        this.entries = entries(this.target);
        this.mem = fromEntries(this.entries.map(([key, value])=>{
            if (typeof value === 'symbol' && pending) {
                if (pending.has(value)) {
                    this.scheduled.add({
                        key,
                        value: pending.get(value)
                    });
                    value = void 0;
                }
            }
            if (typeof value === 'function' && value[FluentCaptureSymbol] === true) {
                this.scheduled.add({
                    key,
                    captured: value._results
                });
                value = void 0;
            }
            return [
                key,
                value
            ];
        }));
    }
    createAccessors() {
        //!? 'creating accessors'
        accessors(this.$, this.target, (key)=>({
                get: ()=>this.mem[key],
                set: (v)=>{
                    //!? 'context setter:', key, v
                    if (v === REDUCER) {
                        v = this.reducer.initial;
                        this.reducer.target = key;
                        // if value is filled first time then return and don't use initial
                        if (this.register(this.reducer)) return;
                        if (v == null) return;
                    }
                    this.target[key] = v;
                }
            }), (key)=>!this.keys.has(key));
        accessors(this.target, this.target, (key)=>({
                get: ()=>this.mem[key],
                set: (v)=>{
                    const prev = this.mem[key];
                    const isDifferent = !Object.is(prev, v);
                    //!? 'context target isDifferent', isDifferent, key, prev, v
                    // if (key === 'presets') console.log('is different', isDifferent, key, v)
                    if (isDifferent) {
                        this.mem[key] = v;
                        this.runEffects(key);
                    // this.maybeAddChangeListener(key, v)
                    }
                }
            }), (key)=>{
            if (!this.keys.has(key)) {
                this.keys.add(key);
                return true;
            }
            return false;
        });
        accessors(this.target, fromEntries(entries(this.mem).filter(([key, value])=>{
            if (this.maybeAddChangeListener(key, value)) return true;
        })), (key)=>({
                set: (v)=>{
                    //!? 'setting event target', key
                    this.maybeAddChangeListener(key, v);
                }
            }));
    }
    cleanup;
    register;
    update(f) {
        let changed = f.pass == null;
        f.pass ??= !f.keys.size;
        for (const key of f.keys){
            const value = this.target[key];
            if (value == null) {
                f.pass = false;
                return true;
            }
            if (!Object.is(f.values[key], value)) {
                f.values[key] = value;
                f.pass = true;
                changed = true;
            }
        }
        return changed;
    }
    flush;
    run;
    get _fulfill() {
        const f = this.active;
        const key = f?.target;
        if (f == null || key == null) {
            throw new TypeError('Attempted to use `fulfill` when no reducer was active.');
        }
        return (value)=>{
            this.target[key] = value;
        };
    }
    finalize() {
        const f = this.active;
        let result = f.fn.call(this, f.values);
        if (f.options?.isFulfillReducer) {
            result = result(this._fulfill);
        }
        if (result?.then) {
            result.then((res)=>{
                try {
                    if (f.cb?.(res) === false) return;
                    if (f.target != null && !f.options?.isFulfillReducer) {
                        this.target[f.target] = res;
                    }
                } finally{
                    if (f.options?.lock) {
                        this.active = null;
                        this.flush();
                    }
                }
            // if (typeof res === 'function') {
            //     res()
            //     f.dispose = null
            // }
            });
            if (!f.options?.lock) {
                this.active = null;
                this.flush();
            }
            return true;
        } else {
            return this.complete(result);
        }
    }
    complete;
    reduce;
    fulfill;
    effect;
    mutate;
    mutating;
    use;
    using;
    function;
    with;
    callback;
    atomic;
    query;
    when;
    withGlobalLock;
    withLock;
    target;
}

const components = new WeakMap();
if (!customElements.getName) {
    const tags = new WeakMap();
    const { define  } = customElements;
    customElements.define = function(name, ctor, options) {
        tags.set(ctor, name);
        return define.call(customElements, name, ctor, options);
    };
    customElements.getName = function(ctor) {
        return tags.get(ctor);
    };
}
let suffixKey = 0;
const fromElement = (ctor, options = ctor.elementOptions ?? {})=>{
    let fn = components.get(ctor);
    if (fn) return fn;
    let tag = customElements.getName(ctor);
    if (!tag) {
        tag = kebab(ctor.name).replace('-element', '');
        if (!tag.includes('-')) tag = 'x-' + tag;
        while(customElements.get(tag) != null){
            tag += (++suffixKey).toString(36);
        }
        customElements.define(tag, ctor, options);
    }
    if (options.extends) fn = (props)=>jsx(options.extends, Object.assign(props, {
            is: tag
        }), void 0);
    else fn = (props)=>jsx(tag, props, void 0);
    components.set(ctor, fn);
    fn.toString = ()=>tag;
    return fn;
};

const AttrTypes = new Map([
    [
        String,
        (x)=>x.toString()
    ],
    [
        Number,
        (x)=>parseFloat(x)
    ],
    [
        Boolean,
        (x)=>x = x === false ? false : x != null
    ],
]);
const applyAttrs = (self, data, initialUpdates)=>{
    for (const [k, v] of Object.entries(data)){
        //!? 'setting attr "%s"', k, v.toString()
        if ([
            String,
            Number,
            Boolean
        ].includes(v)) self[k] = data[k] = void 0;
        else if (v === true) data[k] = true;
        else if (typeof v === 'object') data[k] = v.toString();
    }
    const update = (attr, key, value)=>{
        if (typeof value === 'boolean') {
            if (value) {
                self.setAttribute(attr, '');
            } else {
                if (self.hasAttribute(attr)) {
                    self.removeAttribute(attr);
                } else {
                    data[key] = false;
                }
            }
        } else {
            //!? 'setting', attr, value
            self.setAttribute(attr, value);
        }
    };
    return accessors(self, data, (key)=>{
        const attr = key.toLowerCase();
        return {
            get: ()=>data[key],
            set (value) {
                //!warn 'setter changed', key, value
                if (!self.isConnected) initialUpdates.push(()=>update(attr, key, value));
                else update(attr, key, value);
            }
        };
    });
};
const attrListener = (body)=>new Function('event', `with(this){let fn=${body};return typeof fn=='function'?fn.call(this,event):fn}`);

class PropertySettings {
    attr = false;
    out = false;
    is;
}
const protoPropertyMap = new WeakMap();
const prop = toFluent(PropertySettings, (settings)=>(is)=>(proto, key)=>{
            if (typeof key !== 'string') return;
            settings.is ??= is;
            if (!protoPropertyMap.has(proto)) {
                protoPropertyMap.set(proto, new Map([
                    [
                        key,
                        settings
                    ]
                ]));
            } else {
                const propertyMap = protoPropertyMap.get(proto);
                if (propertyMap.has(key)) {
                    Object.assign(propertyMap.get(key), settings);
                } else {
                    propertyMap.set(key, settings);
                }
            }
        });
const attr = prop.attr;
const out = prop.out;
const is = prop.is;

const RefProxy = (target)=>Getter$1((key)=>({
            get current () {
                return target[key];
            },
            set current (el){
                target[key] = el;
            }
        }));
class Ref extends EventTarget {
    #current;
    key;
    constructor(ref){
        super();
        defineProperty.enumerable.get(()=>this.#current).set((el1)=>{
            this.#current = el1;
            dispatch(this, 'change');
        })(this, 'current');
        Object.assign(this, ref);
    }
}
class PropsRef {
    ref = new Ref();
    constructor(props, key){
        Object.assign(this, props);
        // TODO: implement ref.key in html-vdom
        this.ref.key = key;
    }
}
class RefCollection extends EventTarget {
    items = [];
    listeners = new Map();
    #construct = (param)=>{
        const item = this.construct(param);
        this.listeners.set(item, on(item.ref).change(()=>dispatch(this, 'change')));
        return item;
    };
    [Symbol.iterator]() {
        return this.items[Symbol.iterator]();
    }
    get length() {
        return this.items.length;
    }
    get refs() {
        return filterMap(this.items, (x)=>x.ref.current);
    }
    createItems(entriesOrValues = []) {
        this.items = entriesOrValues.map(this.#construct);
    }
    find(cb) {
        return this.refs.find(cb);
    }
    map(cb) {
        return this.items.map(cb);
    }
    push(state) {
        const result = this.items.push(this.#construct(state));
        dispatch(this, 'change');
        return result;
    }
    insertAfter(state, other) {
        const index = this.refs.indexOf(other);
        if (!~index) {
            throw new ReferenceError('Item not in ref collection.');
        }
        const newItem = this.#construct(state);
        this.splice(index + 1, 0, newItem);
        return newItem;
    }
    splice(start, deleteCount, ...args) {
        const result = this.items.splice(start, deleteCount, ...args);
        dispatch(this, 'change');
        return result;
    }
    move(oldIndex, newIndex) {
        const item = this.items.splice(oldIndex, 1)[0];
        this.splice(newIndex, 0, item);
    }
    add(...items) {
        for (const item of items){
            this.items.push(item);
        }
        dispatch(this, 'change');
    }
    getRefItem(item) {
        const index = this.refs.indexOf(item);
        if (!~index) {
            throw new ReferenceError('Item not in ref collection.');
        }
        return this.items[index];
    }
    remove(item) {
        const index = (item instanceof PropsRef ? this.items : this.refs).indexOf(item);
        if (!~index) {
            throw new ReferenceError('Item not in ref collection.');
        }
        return this.splice(index, 1)[0];
    }
}
class RefMap extends RefCollection {
    #map = new Map();
    construct([key, state]) {
        const item = new PropsRef(state, key);
        this.#map.set(item.ref.key, item);
        return item;
    }
    constructor(entries){
        super();
        this.createItems(entries);
    }
    toJSON() {
        return filterMap(this.items, (x)=>[
                x.ref.key,
                x.ref.current?.toJSON()
            ]);
    }
    get(key) {
        return this.#map.get(key);
    }
}
class RefSet extends RefCollection {
    construct(state) {
        return new PropsRef(state);
    }
    constructor(values){
        super();
        this.createItems(values);
    }
    toJSON() {
        return filterMap(this.items, (x)=>x.ref.current?.toJSON());
    }
}

function element(ctorOrOptions, options) {
    if (typeof ctorOrOptions === 'function') return fromElement(ctorOrOptions, options);
    if (typeof ctorOrOptions === 'object') options = ctorOrOptions;
    return (superclass)=>{
        const parent = superclass;
        // get own prototype decorated properties map
        const ownPropertyMap = protoPropertyMap.get(superclass.prototype) ?? new Map();
        const ownAttrKeys = [
            ...ownPropertyMap
        ].filter(([, settings])=>settings.attr).map(([key])=>key);
        // map from the lowercase (attribute) version to the actual property name
        const attrKeysMap = new Map(ownAttrKeys.map((x)=>[
                x.toLowerCase(),
                x
            ]));
        const attrTypes = new Map();
        // own observed attributes
        const ownObservedAttributes = [
            ...attrKeysMap.keys()
        ];
        //!? 'own observed attributes', ownObservedAttributes, superclass
        const ctor = class extends parent {
            static get elementOptions() {
                return options ?? super.elementOptions;
            }
            static get observedAttributes() {
                return ownObservedAttributes.concat(super.observedAttributes ?? []);
            }
            static get attributeKeys() {
                return ownAttrKeys.concat(super.attributeKeys ?? []);
            }
            static get propertyMap() {
                return new Map([
                    ...ownPropertyMap,
                    ...super.propertyMap ?? []
                ]);
            }
            static get exposedPropertyKeys() {
                return [
                    ...ctor.propertyMap
                ].filter(([, settings])=>settings.out).map(([key])=>key);
            }
            #isMain = true;
            #attrData;
            #attrInitialUpdates = [];
            $;
            host = this;
            isMounted = false;
            constructor(...args){
                super(...args);
                this.#isMain = new.target === ctor;
                if (this.#isMain) {
                    defineProperty.not.enumerable(this, '_ctor', ctor);
                    this.#create();
                    Object.assign(this, args[0]);
                    queueMicrotask(()=>{
                        this.#flushAttributeUpdates();
                    });
                }
            }
            toJSON() {
                return pick(this, ctor.exposedPropertyKeys);
            }
            // goto://context.ts#ContextClass.createMemory
            _scheduleProperty(key, ...args) {
                const symbol = Symbol();
                this._pendingProperties ??= new Map();
                this._pendingProperties.set(symbol, [
                    key,
                    args
                ]);
                return symbol;
            }
            _getPropertySettings(key) {
                return ctor.propertyMap.get(key);
            }
            #applyAttributes() {
                const attrs = pick(this, ctor.attributeKeys);
                for (const [key, value] of entries(attrs)){
                    const valueCtor = value?.constructor;
                    // TODO: maybe turn all as a .is constructor? Boolean needs special case
                    if (valueCtor?.from) {
                        const settings = ctor.propertyMap.get(key);
                        settings.is = valueCtor;
                    }
                    //!? 'value', value
                    const type = value == null ? AttrTypes.get(String) : AttrTypes.get(value) ?? (valueCtor ? AttrTypes.get(valueCtor) ?? (Object.hasOwn(valueCtor.prototype, 'toString') ? AttrTypes.get(String) : null) : null);
                    if (!type) {
                        throw new TypeError(`Attribute "${key}" is not valid type, must be either: String, Number, Boolean, null, undefined`);
                    }
                    //!? 'type', key, type.toString()
                    attrTypes.set(key, type);
                    attrKeysMap.set(key.toLowerCase(), key);
                }
                this.#attrData = attrs;
                applyAttrs(this, attrs, this.#attrInitialUpdates);
            }
            #flushAttributeUpdates() {
                this.#attrInitialUpdates.splice(0).forEach((fn)=>fn());
            }
            #attachReactiveContext() {
                ContextClass.attach(this, omit($, [
                    'transition'
                ]));
                this.host.on = $.on.bind(this, this);
                this.host.dispatch = $.dispatchBind(this);
            }
            #registerChangeEvents() {
                const keys = new Set(ctor.exposedPropertyKeys);
                this.$.effect.keys(keys).task(()=>{
                    $.dispatch.composed(this.host, 'change');
                });
            }
            #create() {
                //!? 'creating', this
                // attach reactive context
                this.#attachReactiveContext();
                this.#applyAttributes();
                // listen for lifecycle events
                $.on(this.host).mounted(()=>{
                    // we flush here as well as at the queue level
                    // because it can be created and mounted before
                    // the next tick, so it will be allowed to create
                    this.#flushAttributeUpdates();
                    // register 'change' events from @$.out() and @$.attr.out()
                    this.#registerChangeEvents();
                });
                $.on(this.host).unmounted(()=>{
                    this.$.cleanup();
                });
                // jsx render helpers
                this.$.render = this.$.effect.cb((result)=>{
                    render(result, this.root ?? (this.root = $.shadow(this)));
                    //!? 'render', this, result
                    return false;
                });
                this.$.part = toFluent(EffectOptions, (options)=>(fn, output)=>{
                        let update;
                        const cb = (value)=>{
                            output = value;
                            if (update) queueMicrotask(update);
                            return false;
                        };
                        const Fn = ()=>{
                            // lazily create effect when first used
                            if (!update) this.$.register(new Fx({
                                fn,
                                cb,
                                options,
                                dispose: cb
                            }));
                            update = hook;
                            return output;
                        };
                        return Fn;
                    });
                this.$.ref = RefProxy(this.$);
                this.created?.(this.$);
            }
            // based on: https://stackoverflow.com/a/49773201/175416
            dispatchEvent(event) {
                const onEventType = `on${event.type}`;
                let pass = true;
                if (this.#isMain && Object.hasOwn(this, onEventType)) {
                    let fn = this[onEventType];
                    if (fn) {
                        if (!fn) fn = attrListener(this.getAttribute(onEventType));
                        pass = fn.call(this, event);
                    }
                }
                // run prototype on<event>(event) listeners
                if (pass !== false) {
                    const fn1 = getOwnProperty(super.constructor.prototype, onEventType);
                    if (fn1) pass = fn1.call(this, event);
                }
                // run prototype <event>(context, event) listeners
                if (pass !== false) {
                    const fn2 = getOwnProperty(super.constructor.prototype, event.type);
                    if (fn2) pass = fn2.call(this, this.$, event);
                }
                if (pass !== false) super.dispatchEvent(event);
                return pass;
            }
            connectedCallback() {
                if (!this.#isMain) {
                    super.connectedCallback?.();
                    return;
                }
                //!? 'connected'
                super.connectedCallback?.();
                if (!this.isMounted) {
                    this.isMounted = true;
                    this.host.dispatch('mounted');
                }
            }
            disconnectedCallback() {
                if (!this.#isMain) {
                    super.disconnectedCallback?.();
                    return;
                }
                super.disconnectedCallback?.();
                // When the element is moved, rather than removed,
                // By observing the state at the end of the task
                // we only fire unmounted when it's been actually
                // immediately.
                queueMicrotask(()=>{
                    if (!this.isConnected) {
                        this.isMounted = false;
                        this.host.dispatch('unmounted');
                    }
                });
            }
            attributeChangedCallback(name, oldValue, newValue) {
                if (!this.#isMain) {
                    super.attributeChangedCallback?.(name, oldValue, newValue);
                    return;
                }
                //!warn 'attributeChangedCallback:', name, newValue
                let key;
                if (key = attrKeysMap.get(name)) {
                    const { is  } = ctor.propertyMap.get(key);
                    const prev = this.#attrData[key];
                    const next = attrTypes.get(key)(newValue);
                    // type casting comparison for objects that serialize their attribute value
                    const isDifferent = is ? prev != next : !Object.is(prev, next);
                    if (isDifferent) {
                        this.#attrData[key] = next;
                        this[key] = is ? this[key] == next ? this[key] : is.from(next) : next;
                        //!? 'propertyChangedCallback:', key, next
                        this.propertyChangedCallback?.(key, prev, next);
                    }
                }
                super.attributeChangedCallback?.(name, oldValue, newValue);
            }
        };
        // name the anonymous class with the decorated class' original name
        defineProperty(ctor, 'name', superclass.name);
        return ctor;
    };
}

class SlottedSettings {
    elements = bool;
    nodes = bool;
    firstChild = bool;
    deep = bool;
}
const slotted$1 = (c)=>toFluent(SlottedSettings, (settings)=>(mapFn = (x)=>x)=>c._scheduleProperty('fulfill', ({ root  })=>(cb)=>$.onSlotChange(root, (slotted)=>{
                        let result;
                        if (settings.nodes) result = slotted.nodes;
                        else if (settings.firstChild) result = [
                            slotted.firstChild
                        ];
                        else result = slotted.elements;
                        if (settings.deep) {
                            result = result.flatMap((x)=>x instanceof HTMLSlotElement ? [
                                    ...x.assignedElements()
                                ] : x);
                        }
                        result = filterMap(result, mapFn);
                        if (!result.length) cb(void 0);
                        else if (settings.firstChild) cb(result[0]);
                        else cb(result);
                    })));

class State extends EventTarget {
    static from(x) {
        return new State(Array.isArray(x) ? x : x.toString().split(','));
    }
    stack;
    constructor(states){
        super();
        this.states = states;
        this.stack = [
            states.Idle
        ];
        $.on(this).change(()=>{
            $.dispatch(this, 'statechange', this);
        });
        if (this.stack.length) {
            $.dispatch(this, 'change');
        }
    }
    toJSON() {
        return this.states;
    }
    get current() {
        return this.toString();
    }
    get isIdle() {
        return this.stack.length === 1;
    }
    toString() {
        return this.stack.at(-1);
    }
    is(testState) {
        return this.stack.at(-1) === testState;
    }
    emit(state, ...detail) {
        if (!this.is(state)) {
            throw new TypeError(`Attempt to emit for "${state}" but our current state stack is "${this.stack}"`);
        }
        $.dispatch(this, state, ...detail);
    }
    push(state, ...detail) {
        if (this.is(state)) {
            throw new TypeError(`Attempt to push state "${state}" again on top of itself, our current state stack is "${this.stack}"`);
        }
        this.stack = [
            ...this.stack,
            state
        ];
        //!warn 'push %s', this.stack
        $.dispatch(this, `${this.stack.at(-2)}pause`);
        $.dispatch(this, `${state}start`, ...detail);
        $.dispatch(this, 'change');
    }
    pop(state, ...detail) {
        if (!this.is(state)) {
            throw new TypeError(`Attempt to pop state "${state}" but our current state stack is "${this.stack}"`);
        }
        if (this.stack.length === 1) {
            throw new TypeError(`Attempt to pop base state "${state}" when stack is length 1.`);
        }
        this.stack = [
            ...this.stack.slice(0, -1)
        ];
        //!warn 'pop %s (-%s)', this.stack, state
        $.dispatch(this, `${state}end`, ...detail);
        $.dispatch(this, `${this.stack.at(-1)}resume`);
        $.dispatch(this, 'change');
    }
    cancel(state, ...detail) {
        if (!this.is(state)) {
            throw new TypeError(`Attempt to cancel state "${state}" but our current state stack is "${this.stack}"`);
        }
        if (this.stack.length === 1) {
            throw new TypeError(`Attempt to cancel base state "${state}" when stack is length 1.`);
        }
        this.stack = [
            ...this.stack.slice(0, -1)
        ];
        //!warn 'cancel %s (-%s)', this.stack, state
        $.dispatch(this, `${state}cancel`, ...detail);
        $.dispatch(this, `${this.stack.at(-1)}resume`);
        $.dispatch(this, 'change');
    }
    swap(state, ...detail) {
        if (this.stack.length === 1) {
            throw new TypeError(`Attempt to swap base state "${state}" when stack is length 1.`);
        }
        const prev = this.current;
        this.stack = [
            ...this.stack.slice(0, -1),
            state
        ];
        $.dispatch(this, `${prev}cancel`);
        $.dispatch(this, `${this.stack.at(-1)}start`, ...detail);
        $.dispatch(this, 'change');
    }
    returnToIdle() {
        while(!this.isIdle){
            this.cancel(this.current, void 0);
        }
    }
    states;
}

/**
 * A finder function that returns first entry that is not nullish.
 *
 * @param e An entry tuple as returned by Object.entries()
 *
 * @private
 */
const NonNull = (e) => e[1] != null;

/**
 * Token interface
 */
class Token extends String {
    static create(value, group, source) {
        return new Token(value, group, source);
    }
    /**
     * The group it matched.
     */
    group;
    /**
     * The value of the match.
     */
    get value() {
        return '' + this;
    }
    /**
     * The index position of the match.
     */
    get index() {
        return this.source.index;
    }
    /**
     * The input string.
     */
    source;
    constructor(value, group, source) {
        super(value);
        this.group = group;
        Object.defineProperty(this, 'source', { enumerable: false, value: source });
    }
    is(group, value) {
        return group == this.group && (value == null || value == this.value);
    }
    as(value, group = this.group) {
        return Token.create(value, group, this.source);
    }
}

/**
 * Convert a match object from `string.match()` or `string.matchAll()`
 * with a named group RegExp such as `/(?<group>[a-z])/` to a more
 * useful {@link Token} object.
 *
 * @param match The match object
 */
const matchToToken = (match) => {
    if (!match)
        return null;
    if (match.index == null)
        return null;
    if (!match.groups)
        throw new TypeError('RegExp match is missing named groups such as: /(?<group>[a-z])/');
    const entry = Object.entries(match.groups).find(NonNull);
    if (entry)
        return Token.create(entry[1], entry[0], match);
    return null;
};

/**
 * Utility to join RegExps together.
 * @private
 */
const joinRegExps = (regexps) => new RegExp(regexps
    .map(x => x.toString())
    .map(x => x.slice(x.indexOf('/') + 1, x.lastIndexOf('/')))
    .join('|'), 'gi');

// eslint-disable-next-line @typescript-eslint/no-unused-vars
/**
 * Create a {@link TokenizerFactory} for the given RegExps.
 *
 * To capture, RegExps must use a [named group](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_Expressions/Groups_and_Ranges#using_named_groups).
 *
 * ```ts
 * const tokenize = createTokenizer(
 *   /(?<ident>[a-z]+)/, // named groups determine token `group`
 *   /(?<number>[0-9]+)/
 * )
 * ```
 *
 * @param regexps RegExps to match.
 */
const createTokenizer = (...regexps) => {
    const regexp = joinRegExps(regexps);
    return ((input) => {
        const matches = input.matchAll(regexp);
        const next = () => matchToToken(matches.next().value);
        const iterator = function* (token) {
            while ((token = next()))
                yield token;
        };
        Object.defineProperty(next, Symbol.iterator, {
            value: iterator,
        });
        return next;
    });
};

const extraWhitespace = /\s+/g;
const tokenizer = createTokenizer(/(?<open>\s*\{\s*)/, /(?<close>\s*\}\s*)/, /(?<comment>\/\*[^]*?\*\/)/, /\s*(?<rule>[^{;}/'"]+)\s+?(?={)/, /\s*(?<string>'.*?'|".*?")(?=;)/, /\s*(?<prop>[^:;/]+)(?=:)/, /\s*(?<value>[^:;]+)(?=;)/);
/**
 * Convert a CSS string to a {@link NestedCSSDeclaration}.
 *
 * @param input
 */ function cssToJs(input) {
    input = input.replace(extraWhitespace, ' ');
    const nextToken = tokenizer(input);
    const filter = (ignored)=>()=>{
            let token;
            while(token = nextToken())if (!ignored.includes(token.group)) break;
            return token;
        };
    const next = filter([
        'comment'
    ]);
    const parse = (style = {})=>{
        let token;
        while(token = next()){
            const { value , group  } = token;
            switch(group){
                case 'rule':
                    next()
                    ;
                    // @ts-ignore can't handle the ??= for some reason
                    parse(style[value.trim()] ??= {});
                    break;
                case 'prop':
                    style[value.trim()] = next()?.value?.trim();
                    break;
                case 'close':
                    return style;
            }
        }
        return style;
    };
    return parse();
}

/**
 * Joins parts with values
 *
 * @param parts
 * @param values
 */ function joinPartsWithValues(parts, values) {
    let str = '';
    for(let i = 0; i < parts.length; i++){
        str += parts[i];
        str += values[i];
    }
    return str;
}
/**
 * Convert string to kebab-case.
 *
 * @param s PascalCase or camelCase string
 */ function kebabCase(s) {
    return s.replace(/[A-Z]/g, (m, i)=>(i ? '-' : '') + m).toLowerCase();
}

const mediaRule = (rule, media)=>media ? `${media}{${rule}}` : rule;
const createRule = (target, prop, value)=>`${target}{${kebabCase(prop)}:${value}}`;
// const isClassOrRoot = /^\s*[.:]/g
const remap = (target, map)=>
    //   ? target
    [
        ...target.matchAll(/[\w-]+|./g)
    ].flat().map((x)=>map?.get(x) ?? x).join('');
/**
 * Compile a JS nested rules {@link NestedCSSDeclaration} to a CSS string.
 *
 * Examples:
 * ```ts
 * jsToCss({ '.foo': { color: 'blue' } })
 * // => .foo{color:blue}
 *
 * // custom root
 * jsToCss({ color: 'red' }, '.my-button')
 * // => .my-button{color:red}
 *
 * // with substitution
 * jsToCss({ '.foo': { color: 'blue' } }, null, new Map([['foo', 'bar']]))
 * // => .bar{color:blue}
 * ```
 *
 * @param rules Rules object
 * @param rootSelector Top level rules will use this selector
 * @param aliasMap Alias identifiers (i.e for `.foo` to become `.bar`
 *  you will need a `foo=bar` entry)
 * @returns The compiled CSS string
 */ function jsToCss(rules, rootSelector, aliasMap) {
    rootSelector ??= '';
    let css = '';
    const parse = (rules, child = '', media = '')=>{
        for(const key in rules){
            const value = rules[key];
            if (typeof value === 'object') {
                const isMedia = /^@/.test(key) ? key : null;
                const isPct = /%$/.test(key);
                let target = child;
                if (isPct) target = key;
                else if (!isMedia) {
                    target = key.split(',').map((k)=>k.trim()).map((k)=>child.split(',').map((x)=>(x + (/^&/.test(k) ? k.slice(1) : ' ' + k)).trim()).join(',')).map((k)=>k.trim()).join(',');
                }
                parse(value, target, isMedia || media);
            } else {
                const rule = mediaRule(createRule(child ? remap(child, aliasMap) : rootSelector, key, value), media);
                css += '\n' + rule.trim();
            }
        }
    };
    parse({
        [rootSelector]: rules
    });
    return css.trim();
}

/**
 * Factory a {@link NestedCSSCompiler} for the given string.
 *
 * @param parts
 * @param values
 */ function css(parts, ...values) {
    const nestedCssString = joinPartsWithValues(parts, values);
    const nestedCssObject = cssToJs(nestedCssString);
    function compileCss(rootSelector = ':host', aliasMap) {
        return jsToCss(nestedCssObject, rootSelector, aliasMap);
    }
    compileCss.valueOf = compileCss;
    return compileCss;
}

/******************************************************************************
Copyright (c) Microsoft Corporation.
Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.
THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}
function __spreadArray(to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
}

/**
 * Helper function to apply arguments to a selected function with type-inference for the supplied function.
 * For instance, in the following statement `applyArg(1).to((x) => x + 1)`, the type of `x` is correctly inferred to be `number`
 * @param args the argument list to be supplied... the argument list of the function supplied to the `to` clause (see below) should match this list
 * @returns an object with a single method, `.to(f)` where `f` is the function to which param `(...args)` will be passed.
 * The return type of this method will match the return type of `f`
 */
var applyArgs = function () {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var f = (function (f) { return f.apply(void 0, __spreadArray([], __read(args), false)); });
    f.to = f;
    return f;
};
function pipeImpl() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var _a = __read(args), o1 = _a[0], operations = _a.slice(1);
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        return operations.reduce(function (acc, f) { return f(acc); }, o1.apply(void 0, __spreadArray([], __read(args), false)));
    };
}
function pipeInto(src, o1) {
    var operations = [];
    for (var _i = 2; _i < arguments.length; _i++) {
        operations[_i - 2] = arguments[_i];
    }
    return applyArgs(src).to(pipeImpl.apply(void 0, __spreadArray([o1], __read(operations), false)));
}

var Intersect;
(function(Intersect) {
    Intersect[Intersect["None"] = 0] = "None";
    Intersect[Intersect["Left"] = 1] = "Left";
    Intersect[Intersect["Top"] = 2] = "Top";
    Intersect[Intersect["Right"] = 4] = "Right";
    Intersect[Intersect["Bottom"] = 8] = "Bottom";
    Intersect[Intersect["Inside"] = 16] = "Inside";
})(Intersect || (Intersect = {}));

class Scalar {
    static interpolate(a, b, t) {
        return a + (b - a) * t;
    }
    static radiansToDegrees(radians) {
        return radians * 180 / Math.PI;
    }
    static degreesToRadians(degrees) {
        return degrees * Math.PI / 180;
    }
}

class Shape {
    x;
    y;
    get left() {
        return this.x;
    }
    set left(x) {
        this.x = x;
    }
    get top() {
        return this.y;
    }
    set top(y) {
        this.y = y;
    }
    get bottom() {
        return this.y;
    }
    set bottom(y) {
        this.y = y;
    }
    get right() {
        return this.x;
    }
    set right(x) {
        this.x = x;
    }
    clone() {
        const ctor = this.constructor;
        return new ctor(this);
    }
    toJSON() {
        return this;
    }
    draw(color = 'red', position = 'absolute') {
        const div = document.createElement('div');
        Object.assign(div.style, {
            position,
            boxSizing: 'border-box',
            border: '1px solid ' + color,
            ...this.toStyle(),
            zIndex: 1000
        });
        document.body.appendChild(div);
        return div;
    }
    screen(other = this) {
        return this.clone().screenSelf(other);
    }
    screenSelf(other = this) {
        return this.translateSelf(other.negate());
    }
    negate() {
        return this.clone().negateSelf();
    }
    negateSelf() {
        this.x = -this.x;
        this.y = -this.y;
        return this;
    }
    contain(other) {
        return this.clone().containSelf(other);
    }
    containSelf(other) {
        if (this.top < other.top) this.top = other.top;
        else if (this.bottom > other.bottom) this.bottom = other.bottom;
        if (this.right > other.right) this.right = other.right;
        else if (this.left < other.left) this.left = other.left;
        return this;
    }
    scale(x = 0, y = x) {
        return this.clone().scaleSelf(x, y);
    }
    scaleSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            this.width *= x.width;
            this.height *= x.height;
        } else {
            this.width *= x;
            this.height *= y;
        }
        return this;
    }
    scaleLinear(x = 0, y = x) {
        return this.clone().scaleLinearSelf(x, y);
    }
    scaleLinearSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            this.width += x.width;
            this.height += x.height;
        } else {
            this.width += x;
            this.height += y;
        }
        return this;
    }
    zoomLinear(x = 0, y = x) {
        return this.clone().zoomLinearSelf(x, y);
    }
    zoomLinearSelf(x = 0, y = x) {
        if (x instanceof Shape) {
            this.scaleLinearSelf(x).translateSelf(x.scale(-0.5));
        } else {
            this.scaleLinearSelf(x, y).translateSelf(x * -0.5, y * -0.5);
        }
        return this;
    }
    add(x = 0, y = x) {
        return this.translate(x, y);
    }
    addSelf(x = 0, y = x) {
        return this.translateSelf(x, y);
    }
    sub(x = 0, y = x) {
        return this.clone().subSelf(x, y);
    }
    subSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            return this.translateSelf(x.negate());
        } else {
            return this.translateSelf(-x, -y);
        }
    }
    translate(x = 0, y = x) {
        return this.clone().translateSelf(x, y);
    }
    translateSelf(x = 0, y = x) {
        if (typeof x === 'object') {
            this.x += x.x || 0;
            this.y += x.y || 0;
        } else {
            this.x += x;
            this.y += y;
        }
        return this;
    }
    touchPoint(other, center = this.center ?? this) {
        const self = this instanceof Rect ? this : new Point(1, 1);
        const w = (self.width + other.width) * 0.5;
        const h = (self.height + other.height) * 0.5;
        const d = center.screen(other.center);
        const tan_phi = h / w;
        const tan_theta = Math.abs(d.y / d.x);
        // tell me in which quadrant the A point is
        const qx = Math.sign(d.x);
        const qy = Math.sign(d.y);
        let xI, yI;
        if (tan_theta > tan_phi) {
            xI = other.center.x + h / tan_theta * qx;
            yI = other.center.y + h * qy;
        } else {
            xI = other.center.x + w * qx;
            yI = other.center.y + w * tan_theta * qy;
        }
        const x = xI - self.width * 0.5;
        const y = yI - self.height * 0.5;
        return new Point(x, y);
    }
    toStylePosition() {
        return {
            left: this.x + 'px',
            top: this.y + 'px'
        };
    }
    toStyleSize() {
        return {
            width: this.width + 'px',
            height: this.height + 'px'
        };
    }
    toStyle() {
        return {
            ...this.toStylePosition(),
            ...this.toStyleSize()
        };
    }
    toPositionObject() {
        return {
            x: this.x,
            y: this.y
        };
    }
    toSizeObject() {
        return {
            width: this.width,
            height: this.height
        };
    }
}

class Point extends Shape {
    static fromElement(el) {
        return new Point(+(el.dataset.left ?? el.offsetLeft), +(el.dataset.top ?? el.offsetTop));
    }
    static fromObject(obj) {
        return new Point(obj.x ?? obj.width, obj.y ?? obj.height);
    }
    static fromMatrix(matrix) {
        return new Point(matrix.e, matrix.f);
    }
    static fromAngle(radians) {
        return new Point(Math.cos(radians), Math.sin(radians));
    }
    static fromAngleDegrees(degrees) {
        return Point.fromAngle(Scalar.degreesToRadians(degrees));
    }
    x;
    y;
    constructor(x = 0, y = x){
        super();
        if (typeof x === 'object') Object.assign(this, Point.fromObject(x.toJSON ? x.toJSON() : x));
        else {
            this.x = x;
            this.y = y;
        }
    }
    [Symbol.iterator]() {
        return [
            this.x,
            this.y
        ][Symbol.iterator]();
    }
    toString() {
        return `${this.x} ${this.y}`;
    }
    set(other) {
        this.x = other.x;
        this.y = other.y;
        return this;
    }
    draw(color = 'yellow', position = 'absolute') {
        const div = document.createElement('div');
        Object.assign(div.style, {
            position,
            boxSizing: 'border-box',
            border: '1px solid ' + color,
            ...this.toStylePosition(),
            width: '2px',
            height: '2px',
            zIndex: 1000
        });
        document.body.appendChild(div);
        return div;
    }
    get pos() {
        return new Point(this.x, this.y);
    }
    get position() {
        return this.pos;
    }
    get size() {
        return new Point(this.width, this.height);
    }
    get width() {
        return this.x;
    }
    set width(x) {
        this.x = x;
    }
    get height() {
        return this.y;
    }
    set height(y) {
        this.y = y;
    }
    interpolate(other, t) {
        return this.clone().interpolateSelf(other, t);
    }
    interpolateSelf(other, t) {
        this.x = this.x + (other.x - this.x) * t;
        this.y = this.y + (other.y - this.y) * t;
        return this;
    }
    diff(other) {
        return this.clone().diffSelf(other);
    }
    diffSelf(other) {
        this.x -= other.x;
        this.y -= other.y;
        return this;
    }
    abs() {
        return this.clone().absSelf();
    }
    absSelf() {
        this.x = Math.abs(this.x);
        this.y = Math.abs(this.y);
        return this;
    }
    square() {
        return this.clone().squareSelf();
    }
    squareSelf() {
        this.x *= this.x;
        this.y *= this.y;
        return this;
    }
    sum() {
        return this.x + this.y;
    }
    manhattan(other) {
        return this.diff(other).absSelf().sum();
    }
    octile(other) {
        const d = this.diff(other).absSelf();
        const F = Math.SQRT2 - 1;
        return d.x < d.y ? F * d.x + d.y : F * d.y + d.x;
    }
    max() {
        return Math.max(this.x, this.y);
    }
    min() {
        return Math.min(this.x, this.y);
    }
    chebyshev(other) {
        return this.diff(other).absSelf().max();
    }
    euclidean(other) {
        return this.distance(other);
    }
    distance(other) {
        return this.diff(other).mag();
    }
    mag() {
        return Math.sqrt(this.square().sum());
    }
    length() {
        return this.mag();
    }
    unit() {
        return this.scale(1 / this.mag());
    }
    dot(other) {
        return new Line(this, other).dot();
    }
    normal() {
        return new Point(this.y, -this.x);
    }
    angleTo(other) {
        return new Line(this, other).angle();
    }
    round() {
        return this.clone().roundSelf();
    }
    roundSelf() {
        this.x = Math.round(this.x);
        this.y = Math.round(this.y);
        return this;
    }
    precisionRound(p = 1) {
        return this.clone().precisionRoundSelf(p);
    }
    precisionRoundSelf(p = 1) {
        this.x = Math.round(this.x * p) / p;
        this.y = Math.round(this.y * p) / p;
        return this;
    }
    gridRound(p = 1) {
        return this.clone().gridRoundSelf(p);
    }
    gridRoundSelf(p = 1) {
        this.x = Math.round(this.x / p) * p;
        this.y = Math.round(this.y / p) * p;
        return this;
    }
    absoluteSum() {
        return Math.abs(this.x) + Math.abs(this.y);
    }
    withinRect(other) {
        return this.x >= other.left && this.x <= other.right && this.y >= other.top && this.y <= other.bottom;
    }
    transform(matrix) {
        return this.clone().transformSelf(matrix);
    }
    transformSelf(matrix) {
        const { x , y  } = this;
        const { a , b , c , d , e , f  } = matrix;
        this.x = a * x + c * y + e;
        this.y = b * x + d * y + f;
        return this;
    }
    multiply(x = 0, y = x) {
        return this.clone().multiplySelf(x, y);
    }
    multiplySelf(x = 0, y = x) {
        if (x instanceof Shape) {
            this.x *= x.x;
            this.y *= x.y;
        } else {
            this.x *= x;
            this.y *= y;
        }
        return this;
    }
    normalize(x, y = x) {
        return this.clone().normalizeSelf(x, y);
    }
    normalizeSelf(x = this.mag(), y = x) {
        if (x instanceof DOMMatrix) {
            this.x = (this.x - x.e) / x.a;
            this.y = (this.y - x.f) / x.d;
        } else if (typeof x === 'number') {
            this.x /= x;
            this.y /= y;
        } else {
            this.x /= x.x;
            this.y /= x.y;
        }
        return this;
    }
    equals(other) {
        return this.x === other.x && this.y === other.y;
    }
    equalsAny(other) {
        return this.x === other.x || this.y === other.y;
    }
}

class Line {
    p1;
    p2;
    constructor(p1, p2){
        if (p1.p1) {
            this.p1 = new Point(p1.p1);
            this.p2 = new Point(p1.p2);
        } else {
            this.p1 = p1;
            this.p2 = p2;
        }
    }
    [Symbol.iterator]() {
        return this.points[Symbol.iterator]();
    }
    get points() {
        return [
            this.p1,
            this.p2
        ];
    }
    angle() {
        const dx = this.p2.x - this.p1.x;
        const dy = this.p2.y - this.p1.y;
        return Math.atan2(dy, dx);
    }
    angleDegrees() {
        return Scalar.radiansToDegrees(this.angle());
    }
    clone() {
        const ctor = this.constructor;
        return new ctor(this);
    }
    mag() {
        return this.p1.distance(this.p2);
    }
    dot() {
        return this.p1.x * this.p2.x + this.p1.y * this.p2.y;
    }
    intersectionRect(r) {
        const p1 = this.p1;
        const p2 = this.p2;
        const is = (this.intersectsLine(r.leftLine) ? Intersect.Left : 0) + (this.intersectsLine(r.topLine) ? Intersect.Top : 0) + (this.intersectsLine(r.rightLine) ? Intersect.Right : 0) + (this.intersectsLine(r.bottomLine) ? Intersect.Bottom : 0) + (p1.withinRect(r) && p2.withinRect(r) ? Intersect.Inside : 0);
        return is;
    }
    intersectsRect(r) {
        const p1 = this.p1;
        const p2 = this.p2;
        // dprint-ignore
        return this.intersectsLine(r.leftLine) ? Intersect.Left : this.intersectsLine(r.topLine) ? Intersect.Top : this.intersectsLine(r.rightLine) ? Intersect.Right : this.intersectsLine(r.bottomLine) ? Intersect.Bottom : p1.withinRect(r) && p2.withinRect(r) ? Intersect.Inside : Intersect.None;
    }
    intersectsLine(other) {
        const a1 = this.p1;
        const a2 = this.p2;
        const b1 = other.p1;
        const b2 = other.p2;
        let q = (a1.y - b1.y) * (b2.x - b1.x) - (a1.x - b1.x) * (b2.y - b1.y);
        const d = (a2.x - a1.x) * (b2.y - b1.y) - (a2.y - a1.y) * (b2.x - b1.x);
        if (d == 0) {
            return false;
        }
        const r = q / d;
        q = (a1.y - b1.y) * (a2.x - a1.x) - (a1.x - b1.x) * (a2.y - a1.y);
        const s = q / d;
        if (r < 0 || r > 1 || s < 0 || s > 1) {
            return false;
        }
        return true;
    }
    // collide line to rectangle and return a new line
    // that is placed just outside of the rectangle and not within
    getLineToRectangleCollisionResponse(intersection, r) {
        if (intersection === Intersect.None) {
            return this;
        }
        const p1 = this.p1;
        const p2 = this.p2;
        const a1 = p1.clone();
        const a2 = p2.clone();
        const b1 = r.topLeft;
        const b2 = r.bottomRight;
        if (intersection & Intersect.Left) {
            a1.x = b1.x - 1;
        }
        if (intersection & Intersect.Top) {
            a1.y = b1.y - 1;
        }
        if (intersection & Intersect.Right) {
            a2.x = b2.x + 1;
        }
        if (intersection & Intersect.Bottom) {
            a2.y = b2.y + 1;
        }
        if (intersection & Intersect.Inside) {
            const p = p1.interpolate(p2, 0.5);
            const tp = p.touchPoint(r);
            a1.x = tp.x;
            a1.y = tp.y;
            a2.x = tp.x;
            a2.y = tp.y;
        }
        //   a1.x = tp.x
        //   a1.y = tp.y
        //   a2.x = tp.x
        //   a2.y = tp.y
        return new Line(a1, a2);
    }
    translate(x = 0, y = x) {
        return this.clone().translateSelf(x, y);
    }
    translateSelf(x = 0, y = x) {
        this.p1.translateSelf(x, y);
        this.p2.translateSelf(x, y);
        return this;
    }
}

class Matrix extends DOMMatrix {
    constructor(matrix){
        if (typeof matrix === 'string') {
            super();
        } else if (Array.isArray(matrix)) {
            super(matrix);
        } else if (typeof matrix === 'object') {
            super(Object.values(matrix));
        } else {
            super(matrix);
        }
    }
    toJSON() {
        return this.toFloat64Array();
    }
}

class Polyline {
    static fromPoints(points) {
        return new Polyline(points.slice(0, -1).map((p, i)=>{
            const next = points[i + 1];
            const line = new Line(p, next);
            return line;
        }));
    }
    lines;
    constructor(lines){
        if (lines.lines) {
            this.lines = lines.lines;
        } else {
            this.lines = lines;
        }
    }
    get normals() {
        const mags = this.lines.map((line)=>line.mag());
        const total = mags.reduce((a, b)=>a + b, 0);
        const normals = mags.map((mag)=>mag / total);
        return normals;
    }
    get path() {
        const points = [];
        for (const line of this.lines){
            points.push(line.p1);
        }
        points.push(this.lines.at(-1).p2);
        return new Polygon(points);
    }
    get length() {
        return this.lines.reduce((a, b)=>a + b.mag(), 0);
    }
    chopAt(index) {
        const points = [];
        for (const [i, line] of this.lines.entries()){
            points.push(line.p1);
            if (i === index) {
                const np = line.p1.interpolate(line.p2, 0.5);
                points.push(np);
            }
        }
        points.push(this.lines.at(-1).p2);
        this.lines = Polyline.fromPoints(points).lines;
        return this;
    }
}

class Rect extends Shape {
    static fromElement(el) {
        return new Rect(+(el.dataset.x ?? el.offsetLeft), +(el.dataset.y ?? el.offsetTop), +(el.dataset.width ?? el.offsetWidth), +(el.dataset.height ?? el.offsetHeight));
    }
    static fromObject(obj) {
        return new Rect(obj.x ?? obj.width, obj.y ?? obj.height, obj.width ?? 0, obj.height ?? 0);
    }
    static fromPoints(topLeft, bottomRight) {
        return new Rect(topLeft.x, topLeft.y, bottomRight.x - topLeft.x, bottomRight.y - topLeft.y);
    }
    static fromUnsortedPoints(p1, p2) {
        const topLeft = new Point(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y));
        const bottomRight = new Point(Math.max(p1.x, p2.x), Math.max(p1.y, p2.y));
        return new Rect(topLeft.x, topLeft.y, bottomRight.x - topLeft.x, bottomRight.y - topLeft.y);
    }
    static combine(rects) {
        const x = Math.min(...rects.map((b)=>b.x));
        const y = Math.min(...rects.map((b)=>b.y));
        return new Rect(x, y, Math.max(...rects.map((b)=>b.right)) - x, Math.max(...rects.map((b)=>b.bottom)) - y);
    }
    x;
    y;
    width;
    height;
    #stale = true;
    constructor(x = 0, y = x, width = x, height = y){
        super();
        if (typeof x === 'object') Object.assign(this, Rect.fromObject(x.toJSON ? x.toJSON() : x));
        else {
            this.x = x;
            this.y = y;
            this.width = width;
            this.height = height;
        }
    }
    [Symbol.iterator]() {
        return [
            this.x,
            this.y,
            this.width,
            this.height
        ][Symbol.iterator]();
    }
    toJSON() {
        return {
            x: this.x,
            y: this.y,
            width: this.width,
            height: this.height
        };
    }
    toString() {
        return `${this.x} ${this.y} ${this.width} ${this.height}`;
    }
    draw(color = 'red', position = 'absolute') {
        const div = document.createElement('div');
        Object.assign(div.style, {
            position,
            boxSizing: 'border-box',
            border: '1px solid ' + color,
            ...this.toStyle(),
            zIndex: 1000
        });
        document.body.appendChild(div);
        return div;
    }
    set(other) {
        this.x = other.x;
        this.y = other.y;
        this.width = other.width;
        this.height = other.height;
        this.#stale = true;
        return this;
    }
    setPosition(other) {
        this.x = other.x;
        this.y = other.y;
        this.#stale = true;
        return this;
    }
    setSize(other) {
        this.width = other.width;
        this.height = other.height;
        this.#stale = true;
        return this;
    }
    get points() {
        return [
            new Point(this.x, this.y),
            new Point(this.right, this.y),
            new Point(this.right, this.bottom),
            new Point(this.x, this.bottom),
        ];
    }
    get pos() {
        return new Point(this.x, this.y);
    }
    get position() {
        return this.pos;
    }
    get size() {
        return new Point(this.width, this.height);
    }
    get center() {
        return new Point(this.x + this.width * 0.5, this.y + this.height * 0.5);
    }
    get topLeft() {
        return this.pos;
    }
    get topRight() {
        return new Point(this.right, this.top);
    }
    get bottomLeft() {
        return new Point(this.x, this.bottom);
    }
    get bottomRight() {
        return this.pos.translate(this.size);
    }
    get right() {
        return this.x + this.width;
    }
    set right(x) {
        this.x = x - this.width;
    }
    get bottom() {
        return this.y + this.height;
    }
    set bottom(y) {
        this.y = y - this.height;
    }
    #leftLine;
    #topLine;
    #rightLine;
    #bottomLine;
    maybeCalculateLines() {
        if (!this.#stale) return;
        this.#leftLine = new Line(this.topLeft, this.bottomLeft);
        this.#topLine = new Line(this.topLeft, this.topRight);
        this.#rightLine = new Line(this.topRight, this.bottomRight);
        this.#bottomLine = new Line(this.bottomLeft, this.bottomRight);
        this.#stale = false;
    }
    get leftLine() {
        this.maybeCalculateLines();
        return this.#leftLine;
    }
    get topLine() {
        this.maybeCalculateLines();
        return this.#topLine;
    }
    get rightLine() {
        this.maybeCalculateLines();
        return this.#rightLine;
    }
    get bottomLine() {
        this.maybeCalculateLines();
        return this.#bottomLine;
    }
    interpolate(other, t) {
        return this.clone().interpolateSelf(other, t);
    }
    interpolateSelf(other, t) {
        this.x = this.x + (other.x - this.x) * t;
        this.y = this.y + (other.y - this.y) * t;
        this.width = this.width + (other.width - this.width) * t;
        this.height = this.height + (other.height - this.height) * t;
        return this;
    }
    round() {
        return this.clone().roundSelf();
    }
    roundSelf() {
        this.x = Math.round(this.x);
        this.y = Math.round(this.y);
        this.width = Math.round(this.width);
        this.height = Math.round(this.height);
        this.#stale = true;
        return this;
    }
    multiply(x = 0, y = x, w = x, h = y) {
        return this.clone().multiplySelf(x, y, w, h);
    }
    multiplySelf(x = 0, y = x, w = x, h = y) {
        if (x instanceof Shape) {
            this.x *= x.x;
            this.y *= x.y;
            this.width *= x.width;
            this.height *= x.height;
        } else {
            this.x *= x;
            this.y *= y;
            this.width *= w;
            this.height *= h;
        }
        this.#stale = true;
        return this;
    }
    transform(matrix) {
        return this.clone().transformSelf(matrix);
    }
    transformSelf(matrix) {
        const { x , y , width , height  } = this;
        const { a , b , c , d , e , f  } = matrix;
        this.x = a * x + c * y + e;
        this.y = b * x + d * y + f;
        this.width = a * width + c * height;
        this.height = b * width + d * height;
        this.#stale = true;
        return this;
    }
    normalize(other) {
        return this.clone().normalizeSelf(other);
    }
    normalizeSelf(other) {
        if (other instanceof DOMMatrix) {
            this.transformSelf(other.inverse());
        } else {
            this.x -= other.left;
            this.y -= other.top;
            this.x /= other.width;
            this.y /= other.height;
            this.width /= other.width;
            this.height /= other.height;
        }
        this.#stale = true;
        return this;
    }
    equals(other) {
        return this.x === other.x && this.y === other.y && this.width === other.width && this.height === other.height;
    }
    intersectionRect(other) {
        return (this.bottom - 0.75 >= other.top ? Intersect.Top : 0) + (this.top + 0.75 <= other.bottom ? Intersect.Bottom : 0) + (this.right - 0.75 >= other.left ? Intersect.Left : 0) + (this.left + 0.75 <= other.right ? Intersect.Right : 0);
    }
    intersectsRect(other) {
        return !(this.bottom - 0.75 < other.top || this.top + 0.75 > other.bottom || this.right - 0.75 < other.left || this.left + 0.75 > other.right);
    }
    withinRect(other) {
        return this.left >= other.left && this.right <= other.right && this.top >= other.top && this.bottom <= other.bottom;
    }
    getCollisionResponse(other) {
        const x = other.x + other.width * 0.5 - (this.x + this.width * 0.5);
        const y = other.y + other.height * 0.5 - (this.y + this.height * 0.5);
        return new Point(x, y);
    }
    place(other, placement) {
        return this.clone().placeSelf(other, placement);
    }
    placeSelf(other, placement) {
        if (placement.includes('n')) this.y = other.top - this.height;
        else if (placement.includes('s')) this.y = other.bottom;
        else this.y = other.center.y - this.height * 0.5;
        if (placement.includes('w')) this.x = other.left - this.width;
        else if (placement.includes('e')) this.x = other.right;
        else this.x = other.center.x - this.width * 0.5;
        if (placement.includes('r')) {
            this.x += this.width;
        }
        if (placement.includes('l')) {
            this.x -= this.width;
        }
        this.#stale = true;
        return this;
    }
}

class Polygon {
    static toSVGPath(points) {
        return (points.length ? points : [
            [
                0,
                0
            ]
        ]).reduce((p, n)=>{
            let last = p.at(-1);
            if (last.length === 3) {
                last = [
                    'L'
                ];
                p.push(last);
            }
            last.push(...n);
            return p;
        }, [
            [
                'M'
            ]
        ]).flat(Infinity).join(' ');
    }
    static sum(points) {
        return points.reduce((p, n)=>{
            p.x += n.x;
            p.y += n.y;
            return p;
        }, new Point());
    }
    static rope(points, coeff = 1) {
        const from = points[0].clone();
        const to = points.at(-1).clone();
        const oldRope = Polyline.fromPoints(points);
        const mags = oldRope.lines.map((line)=>line.mag());
        const total = mags.reduce((a, b)=>a + b, 0);
        const normals = mags.map((mag)=>mag / total);
        const result = points.slice().map((x)=>x.clone());
        const pl = points.length;
        const pl_1 = pl - 1;
        const c = pl_1 / pl;
        let p = to.clone();
        for(let i = pl_1; i--;){
            const t = (i + 1) / pl_1;
            const np = result[i];
            const normal = normals[Math.ceil(i * c)];
            const d = p.screen(np).scale(normal).scale(t ** coeff);
            p.translateSelf(d.negate());
            np.translateSelf(d);
            p = np;
        }
        result[0] = from.clone();
        result[points.length - 1] = to.clone();
        return result;
    }
    static chop(points, min = 35, max = 240) {
        const from = points[0].clone();
        const to = points.at(-1).clone();
        const newRope = Polyline.fromPoints(points);
        const newPoints = [];
        let last;
        for (const line of newRope.lines){
            const mag = line.mag();
            if (mag > min) {
                last = newPoints.push(line.p1);
            } else {
                if (last) {
                    newPoints[last - 1].translateSelf(line.p1.screen(newPoints[last - 1]).scale(0.5));
                }
            }
            if (mag > max) {
                const d = line.p2.screen(line.p1);
                const np = line.p1.translate(d);
                // newPoints.push(line.p1)
                last = newPoints.push(np);
            }
        }
        newPoints[0] = from.clone();
        newPoints.push(to.clone());
        return newPoints;
    }
    static morph(morphFn, from, to, t) {
        const [length, fc, tc] = Morph.coeffs(from, to);
        const fn = morphFn(from, to);
        return t < 1 ? Array.from({
            length
        }, (_, i)=>{
            const fi = Math.round(i * fc);
            const ti = Math.round(i * tc);
            return fn(fi, ti, t);
        }) : to;
    }
    static resample(points, index, t) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const p_0 = points[Math.min(index, pl_1)];
        const p_1 = points[Math.max(0, index - 1)];
        const p$1 = points[Math.min(index + 1, pl_1)];
        return p_0.interpolate(p_1.interpolate(p_0, t).translateSelf(p_0.interpolate(p$1, 1 - t)).scale(0.5), t);
    }
    static fit(points, length) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const coeff = pl / length;
        const newPoints = Array.from({
            length
        }, (_, i)=>{
            const index = Math.round(i * coeff);
            const p_0 = points[Math.min(index, pl_1)];
            const p_1 = points[Math.max(0, index - 1)];
            const p$1 = points[Math.min(index + 1, pl_1)];
            return p_0.interpolate(p_1.interpolate(p_0, 1).translateSelf(p_0.interpolate(p$1, 0)).scale(0.5), 0.5);
        });
        return newPoints;
    }
    // https://www.musicdsp.org/en/latest/Other/60-5-point-spline-interpollation.html
    static resampleSpline(points, index, t) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const p0 = points[Math.max(0, index - 2)];
        const p1 = points[Math.max(0, index - 1)];
        const p2 = points[Math.min(index, pl_1)];
        const p3 = points[Math.min(index + 1, pl_1)];
        const p4 = points[Math.min(index + 2, pl_1)];
        const p5 = points[Math.min(index + 3, pl_1)];
        const x = p2.x + 0.04166666666 * t * ((p3.x - p1.x) * 16.0 + (p0.x - p4.x) * 2.0 + t * ((p3.x + p1.x) * 16.0 - p0.x - p2.x * 30.0 - p4.x + t * (p3.x * 66.0 - p2.x * 70.0 - p4.x * 33.0 + p1.x * 39.0 + p5.x * 7.0 - p0.x * 9.0 + t * (p2.x * 126.0 - p3.x * 124.0 + p4.x * 61.0 - p1.x * 64.0 - p5.x * 12.0 + p0.x * 13.0 + t * ((p3.x - p2.x) * 50.0 + (p1.x - p4.x) * 25.0 + (p5.x - p0.x) * 5.0)))));
        // dprint-ignore
        const y = p2.y + 0.04166666666 * t * ((p3.y - p1.y) * 16.0 + (p0.y - p4.y) * 2.0 + t * ((p3.y + p1.y) * 16.0 - p0.y - p2.y * 30.0 - p4.y + t * (p3.y * 66.0 - p2.y * 70.0 - p4.y * 33.0 + p1.y * 39.0 + p5.y * 7.0 - p0.y * 9.0 + t * (p2.y * 126.0 - p3.y * 124.0 + p4.y * 61.0 - p1.y * 64.0 - p5.y * 12.0 + p0.y * 13.0 + t * ((p3.y - p2.y) * 50.0 + (p1.y - p4.y) * 25.0 + (p5.y - p0.y) * 5.0)))));
        return new Point(x, y);
    }
    // https://www.musicdsp.org/en/latest/Other/49-cubic-interpollation.html
    static resampleCubic(points, index, t) {
        const pl = points.length;
        const pl_1 = pl - 1;
        const i = index + t | 0;
        const p_1 = points[Math.min(Math.max(0, i - 1), pl_1)];
        const p0 = points[Math.min(i, pl_1)];
        const p1 = points[Math.min(i + 1, pl_1)];
        const p2 = points[Math.min(i + 2, pl_1)];
        const ax = (3 * (p0.x - p1.x) - p_1.x + p2.x) * 0.5;
        const bx = 2 * p1.x + p_1.x - (5 * p0.x + p2.x) * 0.5;
        const cx = (p1.x - p_1.x) * 0.5;
        const x = ((ax * t + bx) * t + cx) * t + p0.x;
        const ay = (3 * (p0.y - p1.y) - p_1.y + p2.y) * 0.5;
        const by = 2 * p1.y + p_1.y - (5 * p0.y + p2.y) * 0.5;
        const cy = (p1.y - p_1.y) * 0.5;
        const y = ((ay * t + by) * t + cy) * t + p0.y;
        return new Point(x, y);
    }
    static boundingRect(points) {
        const minX = Math.min(...points.map((p)=>p.x));
        const minY = Math.min(...points.map((p)=>p.y));
        const maxX = Math.max(...points.map((p)=>p.x));
        const maxY = Math.max(...points.map((p)=>p.y));
        return new Rect(minX, minY, maxX - minX, maxY - minY);
    }
    // https://github.com/replit/kaboom/blob/d6985fde12b7073cf7bd300c93e1f79a93453a0a/src/math.ts#L867-L902
    static sat(p1, p2) {
        let overlap = Number.MAX_VALUE;
        let displacement = new Point();
        for (const poly of [
            p1,
            p2
        ]){
            for(let i = 0; i < poly.points.length; i++){
                const a = poly.points[i];
                const b = poly.points[(i + 1) % poly.points.length];
                const axisProj = b.sub(a).normal().unit();
                let min1 = Number.MAX_VALUE;
                let max1 = -Number.MAX_VALUE;
                for(let j = 0; j < p1.points.length; j++){
                    const q = p1.points[j].dot(axisProj);
                    min1 = Math.min(min1, q);
                    max1 = Math.max(max1, q);
                }
                let min2 = Number.MAX_VALUE;
                let max2 = -Number.MAX_VALUE;
                for(let j1 = 0; j1 < p2.points.length; j1++){
                    const q1 = p2.points[j1].dot(axisProj);
                    min2 = Math.min(min2, q1);
                    max2 = Math.max(max2, q1);
                }
                const o = Math.min(max1, max2) - Math.max(min1, min2);
                if (o < 0) {
                    return null;
                }
                if (o < Math.abs(overlap)) {
                    const o1 = max2 - min1;
                    const o2 = min2 - max1;
                    overlap = Math.abs(o1) < Math.abs(o2) ? o1 : o2;
                    displacement = axisProj.scale(overlap);
                }
            }
        }
        return displacement;
    }
    points;
    constructor(polygon = []){
        if (Array.isArray(polygon)) {
            this.points = polygon;
        } else {
            this.points = polygon.points;
        }
    }
}

class Morph {
    static coeffs = (from, to)=>{
        const fl = from.length;
        const tl = to.length;
        const length = Math.max(fl, tl);
        const fc = fl / length;
        const tc = tl / length;
        return [
            length,
            fc,
            tc,
            fl,
            tl
        ];
    };
    static Nearest = (from, to)=>{
        const fl_1 = from.length - 1;
        const tl_1 = to.length - 1;
        return (fi, ti, t)=>{
            const fp = from[Math.min(fi, fl_1)];
            const tp = to[Math.min(ti, tl_1)];
            return fp.interpolate(tp, t);
        };
    };
    static Linear = (from, to)=>(fi, ti, t)=>{
            const fp = Polygon.resample(from, fi, 0.5);
            const tp = Polygon.resample(to, ti, 0.5);
            return fp.interpolate(tp, t);
        };
    static Cubic = (from, to)=>(fi, ti, t)=>{
            const fp = Polygon.resampleCubic(from, fi, 0.5);
            const tp = Polygon.resampleCubic(to, ti, 0.5);
            return fp.interpolate(tp, t);
        };
    static Spline = (from, to)=>(fi, ti, t)=>{
            const fp = Polygon.resampleSpline(from, fi, 0.5);
            const tp = Polygon.resampleSpline(to, ti, 0.5);
            return fp.interpolate(tp, t);
        };
}

// Cardinal spline - a uniform Catmull-Rom spline with a tension option
const cardinal = (data, closed, tension)=>{
    if (data.length < 1) return 'M 0 0';
    if (tension == null) tension = 1;
    const size = data.length - (closed ? 0 : 1);
    const path = [
        `M ${data[0]} C`
    ];
    for(let i = 0; i < size; i++){
        let p0, p1, p2, p3;
        if (closed) {
            p0 = data[(i - 1 + size) % size];
            p1 = data[i];
            p2 = data[(i + 1) % size];
            p3 = data[(i + 2) % size];
        } else {
            p0 = i == 0 ? data[0] : data[i - 1];
            p1 = data[i];
            p2 = data[i + 1];
            p3 = i == size - 1 ? p2 : data[i + 2];
        }
        const x1 = p1.x + (p2.x - p0.x) / 6 * tension;
        const y1 = p1.y + (p2.y - p0.y) / 6 * tension;
        const x2 = p2.x - (p3.x - p1.x) / 6 * tension;
        const y2 = p2.y - (p3.y - p1.y) / 6 * tension;
        path.push(x1, y1, x2, y2, p2.x, p2.y);
    }
    closed && path.push('z');
    return path.join(' ');
};

const isMobileAgent = /Mobi|Android|Tablet/i.test(navigator.userAgent);

const observed = ()=>$.mixin((superclass)=>{
        return class ObservedMixin extends superclass {
            #offsetLeft;
            #offsetTop;
            #offsetWidth;
            #offsetHeight;
            get offsetLeft() {
                return this.#offsetLeft;
            }
            get offsetTop() {
                return this.#offsetTop;
            }
            get offsetWidth() {
                return this.#offsetWidth;
            }
            get offsetHeight() {
                return this.#offsetHeight;
            }
            #updateOffsets() {
                // @ts-ignore
                this.#offsetLeft = super.offsetLeft;
                // @ts-ignore
                this.#offsetTop = super.offsetTop;
                // @ts-ignore
                this.#offsetWidth = super.offsetWidth;
                // @ts-ignore
                this.#offsetHeight = super.offsetHeight;
            }
            rect = new $.Rect();
            pos = $(this).reduce(({ $ , rect  })=>$.pos?.equals(rect.pos) ? $.pos : rect.pos);
            size = $(this).reduce(({ $ , rect  })=>$.size?.equals(rect.size) ? $.size : rect.size);
            get ownRect() {
                this.#updateOffsets();
                const offset = $.getElementOffset(this);
                return new $.Rect(offset.x, offset.y, this.offsetWidth, this.offsetHeight);
            }
            mounted($) {
                $.rect = $.fulfill(({ host  })=>(fulfill)=>{
                        const update = ()=>fulfill(host.ownRect);
                        update();
                        $.observe.resize(host, update);
                    });
            }
        };
    });

var index = /*#__PURE__*/Object.freeze({
    __proto__: null,
    observed: observed
});

const observeNodes = (nodes, observer, options)=>nodes?.forEach((node)=>observer.observe(node, options));
const nodesToText = (nodes)=>nodes?.map((node)=>{
        const text = node.textContent;
        return text.trim().length ? text : '';
    }).join('') ?? '';
const slotted = (slots)=>({
        get nodes () {
            return slots.map((slot)=>slot.assignedNodes()).flat(Infinity);
        },
        get elements () {
            return slots.map((slot)=>slot.assignedElements()).flat(Infinity);
        },
        get firstChild () {
            return this.elements[0];
        }
    });
const onSlotChange = (el, cb, fn = (el)=>el instanceof HTMLSlotElement ? [
        el
    ] : [
        ...el.querySelectorAll('slot')
    ])=>on(el, 'slotchange')(()=>cb(slotted(fn(el))));
const onTextChange = (el, cb, fn)=>{
    let observer;
    const off = onSlotChange(el, ({ nodes  })=>{
        observer?.disconnect();
        if (nodes.length) {
            observer = new MutationObserver(()=>cb(nodesToText(nodes)));
            observeNodes(nodes, observer, {
                characterData: true
            });
            cb(nodesToText(nodes));
        }
    }, fn);
    return ()=>{
        observer?.disconnect();
        return off();
    };
};

const getElementOffset = (el)=>{
    let x = 0;
    let y = 0;
    do {
        x += el.offsetLeft - el.scrollLeft;
        y += el.offsetTop - el.scrollTop;
    }while (el = el.offsetParent)
    return {
        x,
        y
    };
};

const dataset = (target, data)=>{
    for(const key in data)target.dataset[key] = data[key].toString();
};

const findParentElement = (child, fn)=>{
    let el = child;
    while(el = el.parentElement)if (fn(el)) return el;
};

const getActiveElement = (el = document.activeElement)=>{
    if (el?.shadowRoot) return getActiveElement(el.shadowRoot.activeElement) ?? el;
    return el;
};

const kbd = (...shortcuts)=>(e)=>{
        const pressed = [
            e.shiftKey && 'shift',
            e.altKey && 'alt',
            e.ctrlKey && 'ctrl',
            e.metaKey && 'meta',
            (e.ctrlKey || e.metaKey) && 'cmd',
            e.key.replace('Arrow', '').toLowerCase(),
        ].filter(Boolean);
        for (const [shortcut, cmd] of shortcuts){
            if (shortcut.every((x)=>pressed.includes(x.toLowerCase()))) {
                e.preventDefault();
                e.stopPropagation();
                return cmd(e);
            }
        }
    };

const observe = {
    resize: (el, fn)=>{
        const observer = new ResizeObserver(fn);
        observer.observe(el);
        return ()=>observer.disconnect();
    },
    gc: (item, value, fn)=>{
        const reg = new FinalizationRegistry(fn);
        reg.register(item, value);
        return reg;
    }
};

const ondblclick = (el, cb, ms = 250)=>{
    let prevTimestamp = 0;
    const pointers = new Set();
    const clear = (e)=>{
        pointers.delete(e.pointerId);
    };
    return chain(on(el).pointerdown((e)=>{
        if (!(e.buttons & MouseButton.Left)) return;
        pointers.add(e.pointerId);
        if (pointers.size > 1) return;
        if (e.timeStamp - prevTimestamp < ms) {
            e.stopPropagation();
            cb();
            return;
        }
        prevTimestamp = e.timeStamp;
    }), on(window).pointerup(clear), on(window).pointercancel(clear));
};

const shadow = (host, init = '', html = '')=>{
    const root = host.attachShadow(typeof init === 'object' ? init : {
        mode: 'open'
    });
    root.innerHTML = typeof init === 'string' ? init : html;
    return root;
};

class TransitionSettings {
    immediate = bool;
    expire = bool;
    expireAfter = 0;
    locked = bool;
    pop = bool;
    drop = bool;
}
class Transition extends EventTarget {
    static from() {
        throw new TypeError('Transition cannot be recreated by its string representation.');
    }
    #modeExpireTime;
    #dropExpireTime;
    #modeTimeout;
    #resetModeTimeout;
    #locked;
    constructor(state, AnimSettings){
        super();
        this.state = state;
        this.AnimSettings = AnimSettings;
        this.#modeExpireTime = 0;
        this.#dropExpireTime = 0;
        this.#locked = false;
        this.to = toFluent(TransitionSettings, (settings)=>(mode, cb)=>{
                if (this.#locked) {
                    return;
                }
                const now = performance.now();
                if (now < this.#dropExpireTime) {
                    return;
                }
                clearTimeout(this.#resetModeTimeout);
                clearTimeout(this.#modeTimeout);
                if (now > this.#modeExpireTime || settings.immediate) {
                    this.#setMode(settings, mode, cb);
                } else {
                    this.#modeTimeout = setTimeout(()=>{
                        this.#setMode(settings, mode, cb);
                    }, this.#modeExpireTime - now);
                }
            });
        $.on(state).change(()=>{
            this.dispatchEvent(new CustomEvent('change'));
        });
    }
    get animSettings() {
        return this.AnimSettings[this.state.current] ?? this.AnimSettings[this.state.states.Idle];
    }
    #resetMode(cb) {
        clearTimeout(this.#resetModeTimeout);
        this.#resetModeTimeout = setTimeout(()=>{
            requestAnimationFrame(()=>{
                this.#dropExpireTime = 0;
                if (!this.state.isIdle) {
                    this.state.pop(this.state.current);
                }
                if (cb) requestAnimationFrame(cb);
            });
        }, this.AnimSettings[this.state.current]?.duration || 200);
    }
    #setMode(settings, mode, cb1) {
        const now = performance.now();
        if (mode === this.state.states.Idle) this.#resetMode(cb1);
        else {
            this.#modeExpireTime = now + (settings.expireAfter ? settings.expireAfter : settings.expire ? this.AnimSettings[mode]?.duration ?? 0 : 0);
            if (settings.drop) this.#dropExpireTime = this.#modeExpireTime;
            if (settings.locked) this.#locked = true;
            if (mode !== this.state.current) {
                requestAnimationFrame(()=>{
                    if (!this.state.isIdle) {
                        this.state.swap(mode);
                    } else {
                        this.state.push(mode);
                    }
                    if (cb1) requestAnimationFrame(cb1);
                    if (settings.pop) this.#resetMode();
                });
            } else {
                cb1?.();
                if (settings.pop) this.#resetMode();
            }
        }
    }
    unlock() {
        this.#locked = false;
        this.state.pop(this.state.current);
    }
    to;
    get current() {
        return this.toString();
    }
    toString() {
        return this.state.current;
    }
    state;
    AnimSettings;
}
const transition = (state, AnimSettings)=>new Transition(state, AnimSettings);

const warn = (fn)=>function(...args) {
        try {
            return fn.apply(this, args);
        } catch (e) {
            console.warn(e);
        }
    };

function _(ctx) {
    if (typeof ctx === 'function') {
        return ctx;
    } else {
        const c = ctx;
        const shadow = $.shadow.bind(null, ctx);
        const slotted = slotted$1(c);
        return Getter$1((key)=>{
            switch(key){
                case 'shadow':
                    return shadow;
                case 'slotted':
                    return slotted;
                case 'state':
                    return (states)=>new State(states);
                default:
                    {
                        return FluentCapture()[key];
                    }
            }
        });
    }
}
const mixin = (mixin)=>(realsuperclass)=>element()(mixin(realsuperclass));

var sigl = /*#__PURE__*/Object.freeze({
    __proto__: null,
    ContextClass: ContextClass,
    element: element,
    render: render,
    State: State,
    _: _,
    mixin: mixin,
    mix: pipeInto,
    isMobile: isMobileAgent,
    mixins: index,
    AbortOptions: AbortOptions,
    abort: abort,
    MouseButton: MouseButton,
    DispatchOptions: DispatchOptions,
    dispatch: dispatch,
    dispatchBind: dispatchBind,
    EventOptions: EventOptions,
    wrapEvent: wrapEvent,
    event: event,
    chain: chain,
    onAll: onAll,
    OnOptions: OnOptions,
    on: on,
    QueueOptions: QueueOptions,
    wrapQueue: wrapQueue,
    queue: queue,
    Task: Task,
    runTask: runTask,
    groupTasks: groupTasks,
    css: css,
    cssToJs: cssToJs,
    jsToCss: jsToCss,
    get Intersect () { return Intersect; },
    Line: Line,
    Matrix: Matrix,
    Morph: Morph,
    Point: Point,
    Polygon: Polygon,
    Polyline: Polyline,
    Rect: Rect,
    Scalar: Scalar,
    Shape: Shape,
    cardinal: cardinal,
    PropertySettings: PropertySettings,
    protoPropertyMap: protoPropertyMap,
    prop: prop,
    attr: attr,
    out: out,
    is: is,
    RefProxy: RefProxy,
    Ref: Ref,
    PropsRef: PropsRef,
    RefCollection: RefCollection,
    RefMap: RefMap,
    RefSet: RefSet,
    observeNodes: observeNodes,
    nodesToText: nodesToText,
    slotted: slotted,
    onSlotChange: onSlotChange,
    onTextChange: onTextChange,
    getElementOffset: getElementOffset,
    dataset: dataset,
    findParentElement: findParentElement,
    getActiveElement: getActiveElement,
    kbd: kbd,
    observe: observe,
    ondblclick: ondblclick,
    shadow: shadow,
    TransitionSettings: TransitionSettings,
    Transition: Transition,
    transition: transition,
    warn: warn
});

globalThis.Reflect.metadata ??= ()=>{};
const $ = Object.assign(_, sigl, {
    String: String,
    Number: Number,
    Boolean: Boolean
});

/**
 * Copyright (C) 2017-present by Andrea Giammarchi - @WebReflection
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
const {replace} = '';
const ca = /[&<>'"]/g;
const esca = {
  '&': '&amp;',
  '<': '&lt;',
  '>': '&gt;',
  "'": '&#39;',
  '"': '&quot;'
};
const pe = m => esca[m];
/**
 * Safely escape HTML entities such as `&`, `<`, `>`, `"`, and `'`.
 * @param {string} es the input to safely escape
 * @returns {string} the escaped input, and it **throws** an error if
 *  the input type is unexpected, except for boolean and numbers,
 *  converted as string.
 */
const escape = es => replace.call(es, ca, pe);

/**
 * Compiles a syntax definition.
 *
 * ```js
 * const r = await compile({
 *   foo: /[a-z]/,
 *   bar: /[0-9]/,
 * })
 * ```
 *
 * @param def The syntax definition to compile. Can be a promise returned by `import()`.
 */
const compile = async (def, keys = new Set()) => {
    // try to extract the definitions if we're given either module or promise
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const R = (await ((await def?.default) ?? def))?.default ?? def;
    if (!R)
        throw new ReferenceError('Could not find definition for language');
    const map = {};
    const parts = [];
    // eslint-disable-next-line prefer-const
    for (let [name, value] of Object.entries(R)) {
        keys.add(name);
        if (Array.isArray(value)) {
            map[name] = await compile(value[1], keys);
            value = value[0];
        }
        parts.push(`(?<${name}>${value?.source ?? value})`);
    }
    const regexp = RegExp(parts.join('|'), 'gm');
    regexp.map = map;
    regexp.keys = keys;
    return regexp;
};
/**
 * Syntax highlights a string as html with the given syntax.
 *
 * ```js
 * const regexp = await compile({
 *   foo: /[a-z]+/,
 *   bar: [
 *     /[0-9]+/,
 *     {
 *       bar: /[0-5]+/,
 *     },
 *   ],
 * })
 * const html = syntax(regexp, 'hello 123 789 world')
 * ```
 *
 * @param regexp The syntax definition returned by `compile()`
 * @param s The string to highlight.
 */
const syntax = (regexp, s) => s.replace(RegExp('(' + regexp.source + ')|(?<catchbad>[&<>"\'])', 'gm'), (...args) => {
    const groups = args.pop();
    const entries = Object.entries(groups).filter(e => e[1] != null);
    if (!entries.length)
        return '';
    const [tag, value] = entries[0];
    if (tag === 'catchbad')
        return escape(value);
    if (regexp.map) {
        const sub = regexp.map[tag];
        if (sub)
            return `<span class="${tag}-outer">${syntax(sub, value)}</span>`;
    }
    return `<span class="${tag}">${escape(value)}</span>`;
});

const languages = {};
const compiled = new Map();
const style = /*css*/ `
:host {
  contain: layout style paint;
  display: block;
}
* {
  contain: layout style paint;
  text-rendering: optimizeSpeed;
  color: var(--color);
}
*::selection {
  background: var(--selection) !important;
}
pre,
code {
  margin: 0;
  text-rendering: optimizeSpeed;
  font-family: inherit;
}
slot {
  display: none;
}`;
const keyStyle = (x) => /*css*/ `
.${x} {
  color: var(--${x});
  font-style: var(--${x}-style);
  font-weight: var(--${x}-weight);
}`;
let CodeSyntaxElement = class CodeSyntaxElement extends HTMLElement {
    root = $.shadow(this, /*html*/ `<style>${style}</style><slot></slot><pre><code></code></pre>`);
    language = $.String;
    theme = $.String;
    syntax;
    codeRef;
    codeHTML;
    codeRawText;
    highlight;
    onTextContent;
    mounted($) {
        $.codeRef = $.query('code');
        // prevents unstyled content from showing until attributes
        // have arrived and the actual syntax has compiled
        setTimeout(() => {
            if ($.highlight == null)
                $.highlight = escape;
        }, 100);
        $.effect(({ language }) => {
            if (language in languages)
                $.syntax = languages[language];
        });
        $.effect(({ root }) => $.onTextChange(root, text => {
            $.codeRawText = text;
        }));
        $.effect(async ({ root, syntax: syntax$1 }) => {
            let { style, regexp } = compiled.get(syntax$1) ?? {};
            if (!regexp || !style) {
                regexp = await compile(syntax$1);
                style = new CSSStyleSheet();
                await style.replace([...regexp.keys].map(keyStyle).join(''));
                compiled.set(syntax$1, { style, regexp });
            }
            root.adoptedStyleSheets = [style];
            $.highlight = (s) => syntax(regexp, s);
        });
        $.codeHTML = $.reduce(({ codeRawText, highlight }) => highlight(codeRawText));
        $.effect(({ codeRef, codeHTML }) => {
            codeRef.innerHTML = codeHTML;
        });
    }
};
__decorate([
    $.attr(),
    __metadata("design:type", Object)
], CodeSyntaxElement.prototype, "language", void 0);
__decorate([
    $.attr(),
    __metadata("design:type", Object)
], CodeSyntaxElement.prototype, "theme", void 0);
CodeSyntaxElement = __decorate([
    $.element()
], CodeSyntaxElement);

const patchElements = (elements, { theme = 'deep', getLanguage = el => [...el.classList]
    .filter(x => x.startsWith('language-'))[0]
    .split('-')
    .pop(), } = {}) => elements.forEach(oldEl => {
    const newEl = new CodeSyntaxElement();
    newEl.textContent = oldEl.textContent;
    newEl.setAttribute('theme', theme);
    newEl.setAttribute('language', getLanguage(oldEl));
    oldEl.replaceWith(newEl);
});
const patchPreCodeElements = (opts) => patchElements([...document.querySelectorAll('code')].filter(el => el.parentNode?.nodeName === 'PRE'), opts);

export { CodeSyntaxElement, compile, languages, patchElements, patchPreCodeElements, syntax };
//# sourceMappingURL=code-syntax.js.map
